from typing import Optional, Tuple

# Datatypes
DATATYPE_STRING = 'S'
DATATYPE_NUMERIC = 'N'
DATATYPE_BOOLEAN = 'B'
DATATYPE_UNKNOWN = 'U'

# Separator between parts of a compound name
SEPARATOR = '\t'


class Field:
    """
    Class representing a report field. Has a name and a datatype.
    Can be a compound name, where the prefix and suffix parts of
    the name are separated by a tab.
    """
    def __init__(self, name, datatype=DATATYPE_UNKNOWN):
        # Placeholders for computed properties
        self.prefix: Optional[str] = None
        self.suffix: Optional[str] = None

        self.name: str = name
        self.datatype: str = datatype

    def value_of(self, value: str):
        """
        Converts a string into a value of the correct type for
        this field.

        :param value:   The string value to convert.
        :return:        The typed value.
        """
        if self.datatype == DATATYPE_STRING:
            return fix_string(value)
        elif self.datatype == DATATYPE_BOOLEAN:
            return bool(value)
        elif self.datatype == DATATYPE_NUMERIC:
            return float(value)
        else:
            raise ValueError('Cannot convert value as field type is unknown')

    def __setattr__(self, key, value):
        # Validation of datatype
        if key == 'datatype' and not is_valid_datatype(value):
            raise ValueError('datatype must be S, N, B or U')

        super().__setattr__(key, value)

    def __getattribute__(self, item):
        # Lazy calculation of prefix/suffix
        if item == 'prefix' or item == 'suffix':
            self.prefix, self.suffix = split_name(self.name)

        return super().__getattribute__(item)


def split_name(name: str) -> Tuple[Optional[str], Optional[str]]:
    """
    Splits a compound name into its prefix and suffix parts.

    :param name:    TODO
    :return:        TODO
    """
    # If there's no separator, it's not a compound name
    if SEPARATOR not in name:
        return None, None

    # Split around the tab character
    split = name.split(SEPARATOR, 1)

    prefix = split[0]
    suffix = split[1]

    return prefix, suffix


def is_valid_datatype(datatype: str) -> bool:
    """
    Checks if the given datatype string is one of the
    allowable values.

    :param datatype:    The datatype string to check.
    :return:            True if the datatype string is valid,
                        False if not.
    """
    return datatype in {DATATYPE_STRING,
                        DATATYPE_NUMERIC,
                        DATATYPE_BOOLEAN,
                        DATATYPE_UNKNOWN}


def fix_string(string: str) -> str:
    """
    Replaces apostrophes with back-ticks.

    :param string:  The string to fix.
    :return:        The fixed string.
    """
    return string.replace("'", "`")



import inspect
from typing import Optional, Dict, Any

from .typing import AnyCallable


def has_been_overridden(function, obj) -> bool:
    """
    Determines if the given base-class function has been overridden
    by the given object.

    :param function:    A class method (can be bound).
    :param obj:         The object to test for overriding.
    :return:            True if the given object overrides the base-class method,
                        False if not.
    """
    # Unbind the function
    function = unbind(function)

    # Abort if not a function
    if not inspect.isfunction(function):
        raise ValueError(f"{function} is not a function")

    # Get the base class from the function
    function_base_class = get_class_from_function(function)

    # Make sure the function is a class function
    if function_base_class is None:
        raise ValueError("Given function is not a class-function")

    # Get the class of the object
    object_base_class = obj if inspect.isclass(obj) else type(obj)

    # Make sure the object class derives from the base class
    if not issubclass(object_base_class, function_base_class):
        raise ValueError("Given object does not inherit from function base-class")

    # Get the function of the same name from the object base class
    object_base_class_function = unbind(getattr(object_base_class, function.__name__))

    return object_base_class_function is not function


def get_class_from_function(function) -> Optional[type]:
    """
    Gets the class that the given function is defined in.

    :param function:    The function to get the class from.
    :return:            The class, or None if the function is not
                        a class function.
    """
    # If given a bound method, unbind it
    function = unbind(function)

    # Abort if not a function
    if not inspect.isfunction(function):
        return None

    # Get the dotted name of the function
    qual_name = function.__qualname__

    # Split on the dots
    name_parts = qual_name.split(".")

    # If there is only one part, it's not a class function
    if len(name_parts) < 2:
        return None

    # Get the part that should be a class name
    class_name = name_parts[-2]

    # The class should be in the function's global dict
    if class_name not in function.__globals__:
        return None

    # Get the class
    cls = function.__globals__[class_name]

    # Make sure it actually is a class
    if not inspect.isclass(cls):
        return None

    return cls


def is_class_function(function) -> bool:
    """
    Checks if the given function is defined in a class.

    :param function:    The function to check.
    :return:            True if the function is a class function,
                        False if not.
    """
    return get_class_from_function(function) is not None


def unbind(method):
    """
    Unbinds a bound method.

    :param method:  The method to unbind.
    :return:        The unbound function of the given method,
                    or the method itself if it's not bound.
    """
    if not inspect.ismethod(method):
        return method

    return method.__func__


def does_not_raise(func: AnyCallable, *args, **kwargs) -> bool:
    """
    Calls the given function with the given arguments, and reports
    whether it raised an exception or not.

    :param func:    The function to call.
    :param args:    The positional arguments to the function.
    :param kwargs:  The keyword arguments to the function.
    :return:        True if the call completed without raising an exception,
                    False if an exception was raised.
    """
    try:
        func(*args, **kwargs)
    except Exception:
        return False

    return True


def all_as_kwargs(function: AnyCallable, *args, **kwargs) -> Dict[str, Any]:
    """
    Gets a dictionary of the arguments that would be passed to the
    given function by name, resolving positional arguments into
    keyword arguments.

    :param function:    The function.
    :param args:        The positional arguments to the function.
    :param kwargs:      The keyword arguments to the function.
    :return:            All arguments to the function, keyed by parameter name.
                        Includes defaults.
    """
    # Get the function's signature
    signature = inspect.signature(function)

    # Apply the given arguments and defaults
    binding: inspect.BoundArguments = signature.bind(*args, **kwargs)
    binding.apply_defaults()

    return dict(binding.arguments)



from typing import Optional

from ._PythonVersionChecker import PythonVersionChecker


class Python36Checker(PythonVersionChecker):
    """
    Checks if the version of Python is somewhere in 3.6.
    """
    def check(self, major: int, minor: int, micro: int, releaselevel: str, serial: int) -> Optional[str]:
        if major == 3 and minor == 6:
            return None
        return "Not Python 3.6"



from typing import Any, Dict, Union

from ._Field import Field, DATATYPE_BOOLEAN, DATATYPE_STRING

# The pre-defined fields of any report
FIELD_DUMMYREPORT           = 'Dummy report'
FIELD_EXCLUDED              = 'Excluded'
FIELD_SAMPLE_ID             = 'Sample ID'
FIELD_SAMPLE_TYPE           = 'Sample Type'
FIELD_FORMAT                = 'Format'
FIELD_INSERT_TIMESTAMP      = 'Insert Timestamp'
FIELD_INSTRUMENT            = 'Instrument'
FIELD_SOURCE                = 'Source'
FIELD_PARENT_ID             = 'Parent ID'

# The ID# for uninitialised ids
NO_ID = -1


class Report:
    """
    Class representing the header meta-data of a spectrum file.
    """
    def __init__(self):
        self.params: Dict[Field, Any] = {}
        self.fields: Dict[str, Field] = {}
        self.database_id: int = NO_ID

        self.init_fields()

    def init_fields(self):
        """
        Adds the default fields to the report.
        """
        self.fields = {}
        self.add_field(Field(FIELD_DUMMYREPORT, DATATYPE_BOOLEAN))
        self.add_field(Field(FIELD_EXCLUDED, DATATYPE_BOOLEAN))
        self.add_field(Field(FIELD_SAMPLE_ID, DATATYPE_STRING))
        self.add_field(Field(FIELD_SAMPLE_TYPE, DATATYPE_STRING))
        self.add_field(Field(FIELD_INSERT_TIMESTAMP, DATATYPE_STRING))
        self.add_field(Field(FIELD_INSTRUMENT, DATATYPE_STRING))
        self.add_field(Field(FIELD_FORMAT, DATATYPE_STRING))
        self.add_field(Field(FIELD_SOURCE, DATATYPE_STRING))

    def add_field(self, field: Field):
        """
        Adds a new field to the report.

        :param field:   The field to add.
        """
        self.fields[field.name] = field

    def add_parameter(self, field: Union[str, Field], value: any):
        """
        Sets the parameter value for the given field.

        :param field:   The field to set the value against.
        :param value:   The value to apply to the field.
        """
        # Normalise field to its name
        if isinstance(field, Field):
            field_name = field.name
        else:
            field_name = field

        # Can't set the value of a field we don't have
        if field_name not in self.fields:
            raise ValueError('No field called ' + field_name)

        # Get our field reference for the selected field
        field = self.fields[field_name]

        # Convert the value to the field's type
        value = field.value_of(str(value))

        # Set the value against the field
        self.params[field] = value

        # Remember the parent ID as the database ID
        if field_name == FIELD_PARENT_ID:
            self.database_id = value

    def get_parameter(self, field: Union[str, Field]):
        """
        Gets the value set against the given field.

        :param field:   The field to get the value for.
        :return:        The value stored against the field.
        """
        # Normalise a field name to a field reference
        if isinstance(field, str):
            if field in self.fields:
                field = self.fields[field]
            else:
                return None

        # Return the value stored against the field,
        # or None if there isn't one
        if field in self.params:
            return self.params[field]
        else:
            return None

    def set_id(self, id: str):
        """
        Convenience method to set the value of the Sample ID field.

        :param id:  The value to set the Sample ID field to.
        """
        self.add_parameter(FIELD_SAMPLE_ID, id)

    def get_id(self):
        """
        Convenience method to get the value of the Sample ID field.

        :return:    The value of the Sample ID field.
        """
        return self.get_parameter(FIELD_SAMPLE_ID)



from typing import Iterable, Union

from .._typing import OptionsList


def get_options_sub_list(options_list: OptionsList, options: Iterable[str]) -> OptionsList:
    """
    Gets the part of the options list which corresponds to
    the given option/s.

    :param options_list:    The original options list.
    :param options:         The options to extract.
    :return:                The sub-list of options.
    """
    # Get the set of option names
    option_names = set(options)

    # Convert option names to option references
    options = list(getattr(type(self), option) for option in option_names)

    # Convert references to option strings
    option_strings = set(option.option_string for option in options)

    # Get the slices of the options list corresponding to the option
    options_sub_list = []
    slice_started: bool = False
    for string in self._options_list:
        # See if we are in a valid slice
        if string.startswith("--"):
            slice_started = string in option_strings

        # Add the string to the sub-list if it's part of a valid slice
        if slice_started:
            options_sub_list.append(string)

    return options_sub_list



from inspect import isclass, ismethod
from types import MethodType
from typing import Type, TypeVar, Union

# Type variable for the instance type
T = TypeVar("T")


class instanceoptionalmethod:
    """
    Decorator class (like classmethod/staticmethod) which provides
    the class as the first implicit parameter, and the instance as
    the second (if called on an instance) or None (if not).
    """
    def __init__(self, function):
        self._function = function

    def __get__(self, instance, owner):
        # Bind the function to the instance/class
        return MethodType(self._function, instance if instance is not None else owner)

    @staticmethod
    def is_instance(self: Union[T, Type[T]]) -> bool:
        """
        Checks if the given reference is an instance or a class.

        :param self:    The instance/class reference.
        :return:        True if the reference is an instance,
                        False if it is a class.
        """
        return not isclass(self)

    @staticmethod
    def type(self: Union[T, Type[T]]) -> Type[T]:
        """
        Helper method which gets the class from self.

        :param self:    The instance/class reference.
        :return:        The class.
        """
        if instanceoptionalmethod.is_instance(self):
            return type(self)

        return self

    @staticmethod
    def will_fail_on_missing_instance(method: MethodType) -> bool:
        """
        Checks if the given "method" call will fail - i.e. the
        "method" is not actually bound. This is useful for when
        an instance-optional method is overridden by an instance-
        required method, but an attempt may be made to call it
        without an instance.

        :return:    True if the method call will fail because it
                    requires an instance which is not present.
        """
        return not ismethod(method)



import os
from abc import abstractmethod
from typing import IO, Callable, Iterator, Tuple, Union, TypeVar
from io import SEEK_END

from ._FileIOBase import FileIOBase, DiskType, ObjectType

SelfType = TypeVar("SelfType", bound="FileReader")


class FileReader(FileIOBase[DiskType, ObjectType, SelfType]):
    """
    Base class for classes which read files from files. Ensures that
    the file is seekable (e.g. so you can read past the end of a valid
    object in the file and then return the stream position to the start
    of the next object).
    """
    def __init__(self):
        super().__init__()

        self._eof: int = -1
        self._checkpoint: int = -1

    def set_file(self, file: Union[DiskType, IO[DiskType]]) -> SelfType:
        super().set_file(file)

        # Calculate the EOF position
        position: int = self._file.tell()
        self._file.seek(0, SEEK_END)
        self._eof: int = self._file.tell()
        self._file.seek(position)

        return self

    def eof_reached(self) -> bool:
        """
        Returns true if the end-of-file has been reached.
        """
        self.ensure_file()
        return self._file.tell() == self._eof

    def checkpoint(self):
        """
        Sets a checkpoint at the current read position of the file.
        """
        self.ensure_file()
        self._checkpoint = self._file.tell()

    def restore_checkpoint(self):
        """
        Restores the previous checkpoint.
        """
        self.ensure_file()
        self._file.seek(self._checkpoint)

    def load(self) -> ObjectType:
        """
        Loads an object from the file.

        :return:        The loaded object.
        """
        # Make sure we have a file
        self.ensure_file()

        # If reading fails, ensure the file position remains unchanged
        pre_load_position = self._file.tell()
        try:
            return self._load(self._file)
        except Exception:
            self._file.seek(pre_load_position)
            raise

    def loadf(self, filename: str) -> Iterator[ObjectType]:
        """
        Loads all the objects from a file.

        :param filename:    The file to load objects from.
        :return:            The objects.
        """
        with self.set_file(self.open(filename)) as reader:
            while not reader.eof_reached():
                yield reader.load()

    def loads(self, string: DiskType) -> Iterator[ObjectType]:
        """
        Loads all the objects from a string.

        :param string:      The string to load objects from.
        :return:            The objects.
        """
        with self.set_file(string) as reader:
            while not reader.eof_reached():
                yield reader.load()

    def loadd(self,
              path: str,
              filename_predicate: Callable[[str], bool],
              recurse: bool = False) -> Iterator[Tuple[str, int, ObjectType]]:
        """
        Loads all files from the given path that match the filename predicate.

        :param path:                The directory to load from.
        :param filename_predicate:  A function which takes a filename and
                                    returns True if the file should be loaded.
        :param recurse:             Whether sub-directories should be loaded as well.
        :return:                    An iterator over the objects loaded from the files,
                                    along with the filename and the position of the object
                                    in the file.
        """
        # Make sure a valid path was provided
        if not os.path.isdir(path):
            raise ValueError(f"'{path}' is not a directory")

        # Loop through all files/directories in the given directory
        for name in os.listdir(path):
            # Get the full name
            full_name: str = os.path.join(path, name)

            # If we are recursing through directories, do that
            if os.path.isdir(full_name):
                if recurse:
                    for filename, position, obj in self.loadd(full_name, filename_predicate, recurse):
                        yield filename, position, obj

            # Otherwise load the file if the name matches the predicate
            elif os.path.isfile(full_name) and filename_predicate(full_name) is True:
                for position, obj in enumerate(self.loadf(full_name)):
                    yield full_name, position, obj

    @classmethod
    def open(cls, filename: str) -> IO[DiskType]:
        """
        Opens the given file for reading.

        :param filename:    The file to open.
        :return:            The file object.
        """
        return cls._open(filename, 'r')

    @abstractmethod
    def _load(self, file: IO[DiskType]) -> ObjectType:
        """
        Loads a single object from the given file, starting at the current
        file position and leaving the file position at the start of the
        next object (just after the end of the loaded object).

        :param file:    The file to read from. Assumed to be a valid file
                        (readable, seekable).
        :return:        The loaded object.
        """
        pass

    @classmethod
    def _check_file(cls, file: IO[DiskType]):
        super()._check_file(file)

        if not file.readable():
            raise ValueError("File not readable")

        if not file.seekable():
            raise ValueError("File not seekable")



"""
Package for the types of options available to option-handlers.
"""
from ._ClassOption import ClassOption
from ._CountOption import CountOption
from ._FlagOption import FlagOption
from ._Option import Option
from ._TypedOption import TypedOption



from ._AbstractField import AbstractField
from ._DataType import DataType
from ._RegularField import RegularField


class Field(RegularField, AbstractField):
    """
    A single report field identifier.
    """
    @staticmethod
    def is_valid(string: str) -> bool:
        """
        Checks whether the field is valid.

        :param string:  The string to check.
        :return:        True if valid format
        """
        # Must be long enough to have a type-string
        if len(string) <= 3:
            return False

        # Get the type string characters
        string = string[-3:]

        # Make sure the type-string is wrapped in []
        if not string.startswith("[") or not string.endswith("]"):
            return False

        # Get the type-string letter
        string = string[1]

        # Make sure it's a valid type-string letter
        return DataType.parse(string) is not None

    @staticmethod
    def parse_field(string: str) -> "Field":
        """
        Parses the given string and returns the field. The type of the field
        can be append with parentheses: name[type]. Otherwise, UNKNOWN is used
        as type.

        :param string:  The string to parse.
        :return:        The parsed field.
        """
        name: str = string
        datatype: DataType = DataType.UNKNOWN
        if len(string) > 3 and string.endswith("]"):
            type_string: str = string[-3:]
            if type_string.startswith("["):
                type_string = type_string[1]
                datatype = DataType.parse(type_string)
                if datatype is None:
                    datatype = DataType.UNKNOWN
                else:
                    name = name[:-3]

        return Field(Field.unescape(name), datatype)



from typing import IO, Dict, Any

import javaproperties

from wai.common.file._FileIOBase import DiskType, ObjectType
from .._FileReader import FileReader
from ._Report import Report
from . import constants
from ._DataType import DataType
from ._Field import Field


class ReportFileReader(FileReader[str, Report, "ReportFileReader"]):
    """
    Reader for report files.
    """
    def _load(self, file: IO[DiskType]) -> ObjectType:
        # Load the report file as a Java properties file
        properties: Dict[str, Any] = javaproperties.load(file)

        # Parse the properties into a report
        return from_properties(properties)


def from_properties(properties: Dict[str, Any]) -> Report:
    """
    Parses the properties and generates a report object from it.

    :param properties:  The properties to generate the report from.
    :return:            The report.
    """
    # Create the result object
    report: Report = Report()

    # Parse the properties
    for property_name, property_value in properties.items():
        # Data-type properties are only used for the type of their corresponding property
        if property_name.endswith(constants.DATATYPE_SUFFIX):
            continue

        # Special handling for the parent ID field
        if property_name == constants.PROPERTY_PARENTID:
            report.database_id = int(property_value)

        # Get the datatype for this property
        datatype_property_name: str = property_name + constants.DATATYPE_SUFFIX
        datatype_property_value: str = DataType.UNKNOWN.value
        if datatype_property_name in properties:
            datatype_property_value = properties[datatype_property_name]
        datatype: DataType = DataType.parse(datatype_property_value)
        if datatype is None:
            datatype = DataType.UNKNOWN

        # Convert and set the value
        report.set_value(Field(property_name, datatype), datatype.convert(property_value))

    return report



import csv
from typing import IO, Optional, Tuple, List, Type, Union

from .._FileReader import FileReader
from ._CSVFile import CSVFile, DATA_TYPE, VALUE_TYPE, TYPES_TYPE

# Number of lines to sample to try and determine CSV format
SAMPLE_SIZE = 4


class CSVFileReader(FileReader[str, CSVFile, "CSVFileReader"]):
    """
    Reader for CSV files.
    """
    def __init__(self, has_header: Optional[bool] = None, **fmtparams):
        super().__init__()

        self.has_header: Optional[bool] = has_header
        self.fmtparams = fmtparams

    def _load(self, file: IO[str]) -> CSVFile:
        # Sniff the possible dialect
        dialect, has_header = sniff_dialect(file, self.has_header)

        # Get the reader
        reader = csv.reader(file, dialect, **self.fmtparams)

        # Read in all the data
        data = [row for row in reader]

        # Take the header from the data if there is one
        if has_header:
            header = data.pop(0)
        else:
            header = None

        # Get the column types
        types = estimate_types(data[:SAMPLE_SIZE])

        # Convert the data to the given types
        convert_columns(data, types)

        # Create and return the CSV file object
        return CSVFile(header, types, data, dialect)


def sniff_dialect(file: IO[str], has_header: Optional[bool]) -> Tuple[csv.Dialect, bool]:
    """
    Uses the csv Sniffer class to try and auto-detect the dialect of
    the CSV file. Also determines if the file has a header row.

    :param file:        The CSV file to sniff.
    :param has_header:  Whether the file has a header, or None to auto-detect.
    :return:            The detected dialect of the file, and whether it has a header.
    """
    # Peek a small sample of lines from the file
    sample = ''.join([file.readline() for _ in range(SAMPLE_SIZE)])
    file.seek(0)

    # Create the sniffer
    sniffer: csv.Sniffer = csv.Sniffer()

    # Sniff the dialect
    dialect: csv.Dialect = sniffer.sniff(sample)

    # Sniff the header if it isn't already determined
    if has_header is None:
        has_header = sniffer.has_header(sample)

    return dialect, has_header


def estimate_types(sample: DATA_TYPE) -> List[Type[VALUE_TYPE]]:
    """
    Attempts to guess which columns of the provided sample contain numeric data.

    :param sample:          A sample of the data from the CSV file.
    :return:                A list of indices of columns which are thought to contain numeric data.
    """
    # Initialise the list
    types = []

    # Test each column in turn
    for column_index in range(len(sample[0])):
        # Assume the column is integers
        column_type = int

        # Test the value in each row of this column
        for row_index in range(len(sample)):
            # Get the value
            value = sample[row_index][column_index]

            # Test if it converts to an int
            try:
                int(value)
            except Exception:
                # If not, test float
                column_type = float
                try:
                    float(value)
                except Exception:
                    # If not, leave as string
                    column_type = str

        # If all values converted, assume it is a numeric column
        types.append(column_type)

    return types


def convert_columns(data: DATA_TYPE, types: TYPES_TYPE):
    """
    Converts the values in each of the specified columns into the given types.

    :param data:                The data from the CSV file.
    :param types:               The types to convert the data to.
    :return:                    Nothing, the data is converted in place.
    """
    # Convert each row in turn
    for row in data:
        # Convert only those columns that are marked numeric
        for column_index in range(len(row)):
            # Convert the value in place
            row[column_index] = types[column_index](row[column_index])



from .._factorial import factorial


def number_of_subsets(
        set_size: int,
        subset_size: int,
        order_matters: bool = False,
        can_reselect: bool = False
):
    """
    Gets the number of ways to choose ``subset_size`` items from a set of
    ``set_size`` possibilities.

    :param set_size:
                The number of items to select from.
    :param subset_size:
                The number of items to select.
    :param order_matters:
                Whether selections of the same items but in a different selection-order
                are considered distinct subsets.
    :param can_reselect:
                Whether the same item can appear more than once in a subset.
    :return:
                The number of possible subsets that could be selected.
    """
    # Sets can't have negative size
    if set_size < 0:
        raise ArithmeticError(f"Can't have a set of {set_size} items")
    if subset_size < 0:
        raise ArithmeticError(f"Can't have a subset of {subset_size} items")

    # Can only ever select 1 subset of size zero, the empty set
    if subset_size == 0:
        return 1

    # If there are no items to select from, the empty set is the only possible selection,
    # so any subsets of greater size are impossible
    if set_size == 0:
        return 0

    # Handle reselection separately
    if can_reselect:
        # If order matters, (n, k) = n^k
        if order_matters:
            return set_size ** subset_size

        # Otherwise, (n, k) = (n + k - 1, k) (without reselection). Rather than recursing, we
        # just fall through with a modified n
        set_size += subset_size - 1

    else:
        # Without reselection, we can't select more items than are in the set
        if subset_size > set_size:
            return 0

    # If order matters, (n, k) = n! / (n - k)! (without reselection)
    if order_matters:
        return factorial(set_size, set_size - subset_size)

    # Otherwise, (n, k) = n! / k!(n - k)! (again, without reselection).
    # We discriminate on the difference between n and k to determine
    # the least number of multiplications to perform
    remainder_size = set_size - subset_size
    if subset_size > remainder_size:
        return factorial(set_size, subset_size) // factorial(remainder_size)
    else:
        return factorial(set_size, remainder_size) // factorial(subset_size)



from typing import IO

from .._Serialiser import Serialiser


class IntSerialiser(Serialiser[int]):
    def __init__(self,
                 little_endian: bool = True,
                 num_bytes: int = 4,
                 signed: bool = True):
        # Make sure num_bytes is a positive value
        if num_bytes < 1:
            raise ValueError(f"num_bytes must be at least 1")

        self._endianness: str = "little" if little_endian else "big"
        self._num_bytes: int = num_bytes
        self._signed: bool = signed

    def _check(self, obj: int):
        # Make sure the value really is an int
        if not isinstance(obj, int):
            raise TypeError(f"IntSerialiser expects ints, got {type(obj)}")

        # Make sure it is positive if we're in unsigned mode
        if not self._signed and obj < 0:
            raise ValueError(f"Can't serialise negative ints in unsigned mode")

        # Make sure the value is not too big
        if self._bit_length_actual(obj) > 8 * self._num_bytes:
            raise ValueError(f"Int magnitude is too big for current num_bytes setting "
                             f"({obj} won't fit in {self._num_bytes} "
                             f"byte{'s' if self._num_bytes > 1 else ''})")

    def _bit_length_actual(self, obj: int) -> int:
        """
        Gets the actual bit-length of the serialised int,
        taking into account the sign-mode of the serialiser.

        :param obj:     The int to serialise.
        :return:        The serialised bit-length.
        """
        # Zero always takes 1 bit
        if obj == 0:
            return 1

        if self._signed:
            if obj > 0:
                return obj.bit_length() + 1
            else:
                return (obj + 1).bit_length() + 1
        else:
            if obj > 0:
                return obj.bit_length()
            else:
                return -1

    def _serialise(self, obj: int, stream: IO[bytes]):
        stream.write(obj.to_bytes(self._num_bytes, self._endianness, signed=self._signed))

    def _deserialise(self, stream: IO[bytes]) -> int:
        bytes_ = stream.read(self._num_bytes)
        return int.from_bytes(bytes_, self._endianness, signed=self._signed)



from typing import List

from ._number_of_subsets import number_of_subsets


def subset_to_subset_number(
        set_size: int,
        subset: List[int],
        order_matters: bool = False,
        can_reselect: bool = False
) -> int:
    """
    Encodes a subset of the integers in [0, ``set_size``) as a single integer.

    :param set_size:
                The number of items to select from.
    :param subset:
                The selected items by index.
    :param order_matters:
                Whether the selection order of the subset should be encoded as well.
    :param can_reselect:
                Whether duplicate items are allowed in the subset.
    :return:
                A numeric encoding of the subset.
    """
    # Cache the size of the subset
    subset_size = len(subset)

    # Sets can't have negative size
    if set_size < 0:
        raise ArithmeticError(f"'set_size' must be non-negative, got {set_size}")

    # The empty set is the only possible subset of size 0, so encode it as the smallest representation
    if subset_size == 0:
        return 0

    # If there are no items to select from, the empty set is the only possible selection,
    # so any subsets of greater size are impossible
    if set_size == 0:
        raise ArithmeticError(
            f"Can't select a non-empty subset (subset size = {subset_size}) from the empty set"
        )

    # Make sure all values are in [0, n)
    if not all(0 <= x < set_size for x in subset):
        raise ArithmeticError(
            f"Subset contains values outside the valid range for a set-size of {set_size}: {subset}"
        )

    # Handle reselection separately
    if can_reselect:
        # If order matters, shift-encode each value in order
        if order_matters:
            return sum(
                value * (set_size ** index)
                for index, value in enumerate(subset)
            )

        # Otherwise, convert to the equivalent binomial representation and fall-through encode
        counts = {}
        for value in subset:
            counts[value] = counts[value] + 1 if value in counts else 1
        subset = []
        for i in range(set_size - 1):
            last = -1 if len(subset) == 0 else subset[-1]
            subset.append(last + 1 + (0 if i not in counts else counts[i]))
        set_size += subset_size - 1
        subset_size = set_size - subset_size

    else:
        # If not allowed to reselect, subset cannot contain duplicates
        seen = set()
        for value in subset:
            if value in seen:
                raise ArithmeticError(f"Duplicate items in subset: {subset}")
            seen.add(value)

    # If the order matters, shift-encode the items, reducing the problem to (n - 1, k -1)
    # at each iteration
    if order_matters:
        result = subset[0]
        subset = [x - 1 if x > result else x for x in subset[1:]]
        factor = set_size - 1
        while len(subset) > 0:
            result *= factor
            next = subset[0]
            result += next
            factor -= 1
            subset = [x - 1 if x > next else x for x in subset[1:]]

        return result

    # If order doesn't matter, encode the subset as an arithmetic encoding
    sorted = list(subset) if not can_reselect else subset
    sorted.sort(reverse=True)
    return sum(number_of_subsets(n, subset_size - k) for k, n in enumerate(sorted))



from typing import List

from ._number_of_subsets import number_of_subsets


def subset_number_to_subset(
        set_size: int,
        subset_size: int,
        subset_number: int,
        order_matters: bool = False,
        can_reselect: bool = False
) -> List[int]:
    """
    Decodes the result of ``subset_to_subset_number`` into the original subset.

    :param set_size:
                The number of items that were selected from.
    :param subset_size:
                The number of items that were selected.
    :param subset_number:
                The encoded number of the subset, in [0, (n, k)).
    :param order_matters:
                Whether the selection order of the subset was encoded as well.
    :param can_reselect:
                Whether duplicate items were allowed in the subset.
    :return:
                The originally-encoded subset.
    """
    # Sets can't have negative size
    if set_size < 0:
        raise ArithmeticError(f"Can't have a set of {set_size} items")
    if subset_size < 0:
        raise ArithmeticError(f"Can't have a subset of {subset_size} items")

    # Start with the empty set
    subset = []

    # The empty set is the only possible subset of size 0, so return it
    if subset_size == 0:
        # Subset number should be 0 for a subset size of 0
        if subset_number != 0:
            raise ArithmeticError(
                f"0 is the only valid subset number for subsets of size 0, got {subset_number}"
            )

        return subset

    # If there are no items to select from, the empty set is the only possible selection,
    # so any subsets of greater size are impossible
    if set_size == 0:
        raise ArithmeticError(
            f"Can't select a non-empty subset (subset size = {subset_size}) from the empty set"
        )

    # Special case for order-dependent
    if order_matters:
        # Ordered with reselection is shift-encoded, so simply shift-decode
        if can_reselect:
            while len(subset) < subset_size:
                subset.append(subset_number % set_size)
                subset_number //= set_size

        # Without reselection, the items available for selection reduces by 1 at each iteration
        else:
            factor = set_size - subset_size + 1
            while len(subset) < subset_size:
                next = subset_number % factor
                subset_number //= factor
                for index in range(len(subset)):
                    if subset[index] >= next:
                        subset[index] += 1
                subset.insert(0, next)
                factor += 1

        return subset

    # If reselect is allowed, we are expecting the equivalent binomial representation of the selection
    if can_reselect:
        set_size += subset_size - 1
        subset_size = set_size - subset_size

    # Decode the arithmetic encoding of the binomial representation
    num_subsets = number_of_subsets(set_size - 1, subset_size)
    k = subset_size
    for n in reversed(range(set_size)):
        if subset_number >= num_subsets:
            subset_number -= num_subsets
            subset.append(n)
            if len(subset) == subset_size:
                break
            num_subsets = num_subsets * k // n
            k -= 1
        elif n != 0:
            num_subsets = num_subsets * (n - k) // n

    # Convert the binomial representation back to the original multinomial one if reselection was enabled
    if can_reselect:
        subset.sort()
        subset_size = set_size - subset_size
        set_size -= subset_size - 1
        counts = {}
        total = 0
        for i in range(set_size - 1):
            last = -1 if i == 0 else subset[i - 1]
            count = subset[i] - last - 1
            total += count
            if count != 0:
                counts[i] = count
        if total < subset_size:
            counts[set_size - 1] = subset_size - total
        subset = []
        for value, count in counts.items():
            for i in range(count):
                subset.append(value)

    return subset



"""
Module for utilities for dealing with flags.
"""
from typing import Optional

from ..constants import SHORT_FLAG_PREFIX, LONG_FLAG_PREFIX


def is_flag(flag: str) -> bool:
    """
    Checks if the given string is a valid short/long flag.

    :param flag:    The string to check.
    :return:        True if it is a valid flag,
                    False if not.
    """
    return is_short_flag(flag) or is_long_flag(flag)


def is_flag_reason(flag: str) -> Optional[str]:
    """
    Checks if the given string is a valid flag.

    :param flag:    The string to check.
    :return:        None if it is a valid short flag,
                    or the reason if not.
    """
    short_flag_reason = is_short_flag_reason(flag)

    if short_flag_reason is None:
        return None

    long_flag_reason = is_long_flag_reason(flag)

    if long_flag_reason is None:
        return None

    return (f"Not a short flag because: {short_flag_reason}\n"
            f"Not a long flag because: {long_flag_reason}")


def is_short_flag(flag: str) -> bool:
    """
    Checks if the given string is a valid short flag.

    :param flag:    The string to check.
    :return:        True if it is a valid short flag,
                    False if not.
    """
    return is_short_flag_reason(flag) is None


def is_short_flag_reason(flag: str) -> Optional[str]:
    """
    Checks if the given string is a valid short flag.

    :param flag:    The string to check.
    :return:        None if it is a valid short flag,
                    or the reason if not.
    """
    # Short flags must start with '-'
    if not flag.startswith(SHORT_FLAG_PREFIX):
        return f"'{flag}' does not start with '{SHORT_FLAG_PREFIX}'"

    # Must be exactly 2 characters
    if len(flag) > 2:
        return f"'{flag}' is too long (can only be one character after {SHORT_FLAG_PREFIX})"
    elif len(flag) < 2:
        return f"'{flag}' is too short (must be a single letter after {SHORT_FLAG_PREFIX})"

    # The flag character must be a letter
    if not flag[1].isalpha():
        return f"Flag character must be a letter (got '{flag[1]}')"

    return None


def is_long_flag(flag: str) -> bool:
    """
    Checks if the given string is a valid long flag.

    :param flag:    The string to check.
    :return:        True if it is a valid long flag,
                    False if not.
    """
    return is_long_flag_reason(flag) is None


def is_long_flag_reason(flag: str) -> Optional[str]:
    """
    Checks if the given string is a valid long flag.

    :param flag:    The string to check.
    :return:        None if it is a valid long flag,
                    or the reason if not.
    """
    # Long flags must start with '--'
    if not flag.startswith(LONG_FLAG_PREFIX):
        return f"'{flag}' does not start with {LONG_FLAG_PREFIX}"

    # Must have at least 3 characters ('--' followed by at least 1 letter)
    if len(flag) < 3:
        return f"'{flag}' requires at least one character after {LONG_FLAG_PREFIX}"

    # The character after '--' must be a letter
    if not flag[2].isalpha():
        return f"First character after {LONG_FLAG_PREFIX} must be a letter"

    # All other characters must be alpha-numeric or '-'
    if not flag.replace('-', '').isalnum():
        return "All characters must be '-' or alpha-numeric"

    return None


def flag_from_name(name: str) -> str:
    """
    Creates a flag from the given name.

    :param name:    The name, which must be a valid identifier.
    :return:        A flag similar to the name.
    """
    # Name must be an identifier
    if not name.isidentifier() or name.startswith("_"):
        raise ValueError(f"Name ({name}) must be an identifier not starting with an underscore")

    # Create a short flag from single-character names
    if len(name) == 1:
        return SHORT_FLAG_PREFIX + name

    return f"--{name.replace('_', '-')}"



from typing import Any

from ._AbstractField import AbstractField
from ._Field import Field
from ._PrefixOnlyField import PrefixOnlyField
from ._DataType import DataType


class PrefixField(Field, PrefixOnlyField):
    """
    A compound field that only displays the first half of the name.
    """
    # The dummy suffix
    DUMMY_SUFFIX: str = "DUMMY"

    def __init__(self, *args):
        if len(args) == 0:
            args = (None, DataType.NUMERIC)
        elif len(args) > 2:
            raise ValueError("PrefixField only takes 0, 1 or 2 arguments to __init__")

        super().__init__(*args)

        if self.name is not None:
            if not self.is_compound():
                self.name = self.name + self.SEPARATOR + self.DUMMY_SUFFIX
            else:
                self.name = self.get_prefix() + self.SEPARATOR + self.DUMMY_SUFFIX

    def compare_to(self, other: Any) -> int:
        if other is None:
            return 1

        if not isinstance(other, AbstractField):
            return -1

        if self.get_prefix() is None and other.get_prefix() is None:
            return 0

        if self.get_prefix() is None:
            return -1

        if other.get_prefix() is None:
            return 1

        return (-1 if self.get_prefix().lower() < other.get_prefix().lower() else
                0 if self.get_prefix().lower() == other.get_prefix().lower() else
                1)

    def to_string(self) -> str:
        if self.name is None:
            return ""

        return self.get_prefix()

    def to_display_string(self) -> str:
        return self.to_string()

    @staticmethod
    def parse_field(string: str) -> Field:
        return PrefixField(Field.parse_field(string))



"""
Module containing types to do with option parsing.
"""
from typing import List, Any

# The type of a list of options
OptionsList = List[str]


def is_options_list(options_list: Any) -> bool:
    """
    Checks whether the given value is an options list.

    :param options_list:    The value to check.
    :return:                True if it is an options list.
    """
    return (
        isinstance(options_list, list) and
        all(isinstance(element, str) for element in options_list)
    )



from typing import Generic, TypeVar, IO

from .._Serialiser import Serialiser

TupleType = TypeVar("TupleType", bound=tuple)


class TupleSerialiser(
    Serialiser[TupleType],
    Generic[TupleType]
):
    def __init__(
            self,
            *element_serialisers: Serialiser  # Should be a serialiser matching each element of TupleType,
                                              # but this is not supported yet
    ):
        super().__init__()

        self._element_serialisers = element_serialisers

    def _check(
            self,
            obj: TupleType
    ):
        # Must be a dictionary
        if not isinstance(obj, tuple):
            raise ValueError(f"{TupleSerialiser.__class__.__name__} serialises tuples")

        # Must have the same number of elements as we have element-serialisers
        num_required_elements = len(self._element_serialisers)
        num_received_elements = len(obj)
        if num_received_elements != num_required_elements:
            raise ValueError(f"Expected tuple of {num_required_elements} elements but received {num_received_elements}")

    def _serialise(
            self,
            obj: TupleType,
            stream: IO[bytes]
    ):
        # Serialise each value in turn
        for value, serialiser in zip(obj, self._element_serialisers):
            serialiser.serialise(value, stream)

    def _deserialise(
            self,
            stream: IO[bytes]
    ) -> TupleType:
        # Deserialise the elements in serialised order
        return tuple(
            serialiser.deserialise(stream)
            for serialiser in self._element_serialisers
        )



from typing import Tuple, Dict, Any, Set

from .constants import ABSTRACT_METHOD_ATTRIBUTE, ABSTRACT_CLASS_ATTRIBUTE


def get_abstract_methods(bases: Tuple, namespace: Dict[str, Any]):
    """
    Gets the names of the abstract methods that will result from
    a class created with the given base-class set and namespace.

    :param bases:       The base-class set.
    :param namespace:   The namespace.
    :return:            The set of abstract method names.
    """
    abstract_methods = set()

    for base in bases:
        abstract_methods.update(abstract_methods_of(base))

    for name, value in namespace.items():
        if is_abstract_function(value):
            abstract_methods.add(name)
        elif name in abstract_methods:
            abstract_methods.remove(name)

    return abstract_methods


def will_be_abstract(bases: Tuple, namespace: Dict[str, Any]):
    """
    Determines if a class made with the given set of base-classes
    and namespace will be abstract or concrete.

    :param bases:       The set of base classes.
    :param namespace:   The namespace of the new class.
    :return:            True if the newly-created class will be abstract,
                        False if it will be concrete.
    """
    return len(get_abstract_methods(bases, namespace)) > 0


def is_abstract_function(func):
    """
    Whether the given function is abstract in its class.

    :param func:    The function to check.
    :return:        True if the function is abstract,
                    False if not.
    """
    return getattr(func, ABSTRACT_METHOD_ATTRIBUTE, False)


def abstract_methods_of(cls) -> Set[str]:
    """
    Gets the abstract methods of a class.

    :param cls:     The class to get the abstract methods from.
    :return:
    """
    return getattr(cls, ABSTRACT_CLASS_ATTRIBUTE, set())


def is_abstract_class(cls) -> bool:
    """
    Checks if the given class is abstract.

    :param cls:     The class to check.
    :return:        True if the given class is abstract,
                    False if not.
    """
    return len(abstract_methods_of(cls)) > 0



from collections import Hashable
from typing import Any


def is_hashable(obj: Any) -> bool:
    """
    Checks if the given object is hashable.

    :param obj: The object to check.
    :return:    True if the object is hashable,
                False if not.
    """
    return isinstance(obj, Hashable)



from typing import IO

from .._Serialiser import Serialiser
from ._IntSerialiser import IntSerialiser


class LengthPrefixedBytesSerialiser(Serialiser[bytes]):
    """
    Serialiser which serialises raw binary data. Inserts the length of
    the data in bytes before the raw binary data.
    """
    def __init__(
            self,
            length_serialiser: Serialiser[int] = IntSerialiser(signed=False)
    ):
        super().__init__()
        self._length_serialiser: Serialiser[int] = length_serialiser

    def _check(self, obj: bytes):
        # All binary data is valid
        pass

    def _serialise(self, obj: bytes, stream: IO[bytes]):
        # Serialise the length of the data
        self._length_serialiser.serialise(len(obj), stream)

        # Write the data
        stream.write(obj)

    def _deserialise(self, stream: IO[bytes]) -> bytes:
        # Read the length from the stream
        byte_length: int = self._length_serialiser.deserialise(stream)

        return stream.read(byte_length)



from typing import Generic, TypeVar, Type, Union, Optional, Any

import typing_inspect

# The type of the forward-direction keys for the two-way dict
FORWARD_TYPE = TypeVar("FORWARD_TYPE")

# The type of the reverse-direction keys for the two-way dict
REVERSE_TYPE = TypeVar("REVERSE_TYPE")

# Type that applies to either key
EITHER_TYPE = Union[FORWARD_TYPE, REVERSE_TYPE]


class TwoWayDict(Generic[FORWARD_TYPE, REVERSE_TYPE]):
    """
    A two-way dictionary, which can be accessed by either key or value.
    Attempts to auto-discriminate which direction a method call should be
    applied in (forward or reverse) based on the argument types. If the
    keys and values are of the same type, this must be specified explicitly
    with the method with the required suffix.
    """
    def __init__(self):
        self._forward_map = dict()
        self._reverse_map = dict()

    def forward_type(self) -> Type[FORWARD_TYPE]:
        """
        Gets the type of the forward-direction keys for this two-way dict.
        """
        return typing_inspect.get_args(self.__orig_class__)[0]

    def reverse_type(self) -> Type[REVERSE_TYPE]:
        """
        Gets the type of the reverse-direction keys for this two-way dict.
        """
        return typing_inspect.get_args(self.__orig_class__)[1]

    def matches_forward_type(self, value: Any, only: bool = False) -> bool:
        """
        Whether the type of the given key matches the forward type.

        :param value:   The key to check.
        :param only:    Whether the value is not also allowed to match the
                        reverse type.
        :return:        True if matches, False if not.
        """
        forward_type: Type[FORWARD_TYPE] = self.forward_type()

        if typing_inspect.get_origin(forward_type) is type:
            if not isinstance(value, type):
                return False
            forward_type = typing_inspect.get_args(forward_type)[0]
            check = issubclass
        else:
            check = isinstance

        matches: bool = check(value, forward_type)

        if matches and only:
            matches = matches and not self.matches_reverse_type(value)

        return matches

    def matches_reverse_type(self, value: Any, only: bool = False) -> bool:
        """
        Whether the type of the given key matches the reverse type.

        :param value:   The key to check.
        :param only:    Whether the value is not also allowed to match the
                        reverse type.
        :return:        True if matches, False if not.
        """
        reverse_type: Type[REVERSE_TYPE] = self.reverse_type()

        if typing_inspect.get_origin(reverse_type) is type:
            if not isinstance(value, type):
                return False
            reverse_type = typing_inspect.get_args(reverse_type)[0]
            check = issubclass
        else:
            check = isinstance

        matches: bool = check(value, reverse_type)

        if matches and only:
            matches = matches and not self.matches_forward_type(value)

        return matches

    def both_types_same(self) -> bool:
        """
        Returns whether the type of the forward-direction keys is the same
        as the reverse-direction keys.
        """
        return self.forward_type() is self.reverse_type()

    def set(self, key: EITHER_TYPE, value: EITHER_TYPE):
        """
        Sets an entry in this two-way dict. Automatically infers the
        direction based on key/value types.

        :param key:     The key to set.
        :param value:   The value to set against it.
        """
        return self.type_select(key, value,
                                forward_result=self.set_forward,
                                reverse_result=self.set_reverse)(key, value)

    def __setitem__(self, key: EITHER_TYPE, value: EITHER_TYPE):
        """
        Sets an entry in this two-way dict. Automatically infers the
        direction based on key/value types.

        :param key:     The key to set.
        :param value:   The value to set against it.
        """
        self.set(key, value)

    def set_forward(self, key: FORWARD_TYPE, value: REVERSE_TYPE):
        """
        Explicitly sets a mapping in the forward direction.

        :param key:     The forward-direction key.
        :param value:   The value.
        """
        if key in self._forward_map:
            self._reverse_map.pop(self._forward_map[key])

        if value in self._reverse_map:
            self._forward_map.pop(self._reverse_map[value])

        self._forward_map[key] = value
        self._reverse_map[value] = key

    def set_reverse(self, key: REVERSE_TYPE, value: FORWARD_TYPE):
        """
        Explicitly sets a mapping in the reverse direction.

        :param key:     The reverse-direction key.
        :param value:   The value.
        """
        self.set_forward(value, key)

    def get(self, key: EITHER_TYPE) -> EITHER_TYPE:
        """
        Gets the value of an entry in this two-way dict. Automatically infers the
        direction based on key type.

        :param key:     The key to get the value for.
        """
        return self.type_select(key,
                                forward_result=self.get_forward,
                                reverse_result=self.get_reverse)(key)

    def __getitem__(self, item: EITHER_TYPE) -> EITHER_TYPE:
        """
        Gets the value of an entry in this two-way dict. Automatically infers the
        direction based on key type.

        :param item:     The key to get the value for.
        """
        return self.get(item)

    def get_forward(self, key: FORWARD_TYPE) -> REVERSE_TYPE:
        """
        Explicitly gets the value of a mapping in the forward direction.

        :param key:     The forward-direction key.
        :return:        The corresponding value.
        """
        return self._forward_map[key]

    def get_reverse(self, key: REVERSE_TYPE) -> FORWARD_TYPE:
        """
        Explicitly gets the value of a mapping in the reverse direction.

        :param key:     The reverse-direction key.
        :return:        The corresponding value.
        """
        return self._reverse_map[key]

    def contains(self, key: EITHER_TYPE) -> bool:
        """
        Checks if this two-way dict contains the given key. Automatically
        infers the direction based on the key's type.

        :param key:     The key to check for.
        :return:        True if the key is in the dict, False if not.
        """
        return self.type_select(key,
                                forward_result=self.contains_forward,
                                reverse_result=self.contains_reverse)(key)

    def __contains__(self, item: EITHER_TYPE) -> bool:
        """
        Checks if this two-way dict contains the given key. Automatically
        infers the direction based on the key's type.

        :param item:    The key to check for.
        :return:        True if the key is in the dict, False if not.
        """
        return self.contains(item)

    def contains_forward(self, key: FORWARD_TYPE) -> bool:
        """
        Explicitly checks if the given key is contained in the
        dict in the forward direction.

        :param key:     The key to check for.
        :return:        True if the key is in the forward-direction of the dict,
                        False if not.
        """
        return key in self._forward_map

    def contains_reverse(self, key: REVERSE_TYPE) -> bool:
        """
        Explicitly checks if the given key is contained in the
        dict in the reverse direction.

        :param key:     The key to check for.
        :return:        True if the key is in the reverse-direction of the dict,
                        False if not.
        """
        return key in self._reverse_map

    def __iter__(self):
        """
        Iterates through all pairs of keys in the dict.
        """
        return (pair for pair in self._forward_map.items())

    def __len__(self) -> int:
        """
        The number of key-pairs in the dict.
        """
        return len(self._forward_map)

    def type_select(self, key: EITHER_TYPE, value: Optional[EITHER_TYPE] = None, *,
                    forward_result, reverse_result):
        """
        Automatically selects between a forward and reverse direction result based
        on the types of the key and value given.

        :param key:             The key.
        :param value:           The value.
        :param forward_result:  The result to return if the forward direction is inferred.
        :param reverse_result:  The result to return if the reverse direction is inferred.
        :return:                One of the two results.
        """
        if self.both_types_same():
            raise TypeError("Can't auto-differentiate direction when both types the same")

        if value is not None:
            if self.matches_forward_type(key) and self.matches_reverse_type(value):
                return forward_result
            elif self.matches_reverse_type(key) and self.matches_forward_type(value):
                return reverse_result
            else:
                raise TypeError("Key and value types do not align with dict types")
        else:
            if self.matches_forward_type(key, True):
                return forward_result
            elif self.matches_reverse_type(key, True):
                return reverse_result
            else:
                raise TypeError("Key type does not align with dict types")



"""
Module containing error types for when problems occur during
handling of code representations.
"""
from typing import Any


class CodeRepresentationError(Exception):
    """
    Type of error occurring when there is a problem getting the
    code-representation of a value.
    """
    pass


class IsNotCodeRepresentableValue(CodeRepresentationError):
    """
    Error for when a value is not code-representable.
    """
    def __init__(self, value: Any):
        super().__init__(f"Value '{value}' is not code-representable")


class IsNotCodeRepresentableType(CodeRepresentationError):
    """
    Error for when a type is not code-representable.
    """
    def __init__(self, cls: type):
        super().__init__(f"Type '{cls.__qualname__}' is not code-representable")


class ConflictingImports(CodeRepresentationError):
    """
    Error for when combining import dictionaries and there is
    more than one import under a given identifier.
    """
    def __init__(self, identifier: str, import_code_1: str, import_code_2: str):
        super().__init__(f"Multiple conflicting imports with name '{identifier}'\n"
                         f"{import_code_1}\n"
                         f"<->\n"
                         f"{import_code_2}")



from typing import IO

import javaproperties

from .._FileWriter import FileWriter
from ._Report import Report
from . import constants


class ReportFileWriter(FileWriter[str, Report, "ReportFileWriter"]):
    """
    Writer for report files.
    """
    def _dump(self, obj: Report, file: IO[str]):
        # Create a Java properties dictionary
        properties = {constants.PROPERTY_PARENTID: str(obj.database_id)}

        # Add the fields and their types as properties
        for field in obj:
            name = str(field)
            properties[name] = str(obj.get_value(field))
            properties[name + constants.DATATYPE_SUFFIX] = field.datatype.to_display()

        print(javaproperties.to_comment(javaproperties.java_timestamp()), file=file)
        print(javaproperties.to_comment("Simple report format (= Java properties file format)"), file=file)
        javaproperties.dump(properties, file, timestamp=False, sort_keys=True)



from ._ClassRegistry import ClassRegistry
from ._functions import is_hashable
from ._Interval import Interval
from ._InvalidStateError import InvalidStateError
from ._memoise import Memoiser, MemoiserFactory, MEMO_EXTENSION, MemoFile, InvalidMemoFileError, PicklableDict
from ._pool import run_on_all, num_processes
from ._switch import switch, case, default, break_
from ._TwoWayDict import TwoWayDict



"""
Module for constant values related to CLI functionality.
"""
from typing import Set as _Set

# The quote characters
QUOTES: _Set[str] = {"'", '"'}

# The prefix characters for a short flag
SHORT_FLAG_PREFIX: str = '-'

# The prefix characters for a long flag
LONG_FLAG_PREFIX: str = '--'



"""
Package for logging utilities.
"""
from ._LoggingMixin import LoggingMixin
from ._root_logger import root_logger
from ._standard_application_root_logger import (
    DEBUG_HANDLER_NAME,
    INFO_HANDLER_NAME,
    WARNING_HANDLER_NAME,
    ERROR_HANDLER_NAME,
    exact_level_filter,
    create_standard_handler,
    create_standard_application_root_logger
)
from ._standard_library_root_logger import create_standard_library_root_logger



from numbers import Real
from typing import Any, Dict, List, Union, Optional

from . import constants
from ._DataType import DataType
from ._Field import Field
from ._AbstractField import AbstractField
from ._PrefixOnlyField import PrefixOnlyField
from ._SuffixOnlyField import SuffixOnlyField
from ._PrefixField import PrefixField
from ._SuffixField import SuffixField


class Report:
    """
    Python implementation of the Report class in the adams-base -> adams-core -> adams.data.report package.
    """
    def __init__(self):
        # Store Header parameters ( parameter:value )
        self._params: Dict[AbstractField, Any] = {}

        # Fields
        self._fields: Dict[str, AbstractField] = {}

        # The database ID of the data structure this report belongs to
        self.database_id: int = constants.NO_ID

        self.init_fields()

    def init_fields(self):
        """
        Set field types.
        """
        self._fields = {}

        self.add_field(Field(constants.FIELD_DUMMYREPORT, DataType.BOOLEAN))
        self.add_field(Field(constants.FIELD_EXCLUDED, DataType.BOOLEAN))

    def add_field(self, field: AbstractField):
        """
        Adds the given field.

        :param field:   The field to add.
        """
        self._fields[field.name] = field

    def has_field(self, field: AbstractField) -> bool:
        """
        Checks whether the field is already stored.

        :param field:   The field to check.
        :return:        True if already added.
        """
        return field.name in self._fields

    def get_field_type(self, field: AbstractField) -> DataType:
        """
        Returns the type for the given field. Either based on the stored
        fields or DataType.UNKNOWN.

        :param field:   The field to look for.
        :return:        The stored field type or DataType.UNKNOWN if field not stored.
        """
        if self.has_field(field):
            return self._fields[field.name].datatype

        return DataType.UNKNOWN

    def get_fields(self, fix: Union[PrefixOnlyField, SuffixOnlyField, None] = None) -> List[AbstractField]:
        """
        Gets all fields as a list.

        :param fix: The common prefix/suffix to match, or none to get all.
        :return:    The fields.
        """
        # Decide the match criteria for the fields
        if isinstance(fix, PrefixOnlyField):
            to_match = fix.get_prefix()

            def field_matcher(field: AbstractField) -> bool:
                return field.is_compound() and field.get_prefix() == to_match
        elif isinstance(fix, SuffixOnlyField):
            to_match = fix.get_suffix()

            def field_matcher(field: AbstractField) -> bool:
                return field.is_compound() and field.get_suffix() == to_match
        else:
            def field_matcher(field: AbstractField) -> bool:
                return True

        return [field for field in self._params.keys() if field_matcher(field)]

    def get_prefix_fields(self) -> List[PrefixOnlyField]:
        """
        Returns all distinct prefix fields.

        :return:    The prefix fields.
        """
        # Get the unique prefix fields as a set
        result = {PrefixField(field) for field in self._params.keys() if field.is_compound()}

        # Put the prefix fields into a sorted list
        result = list(result)
        result.sort()

        return result

    def get_suffix_fields(self) -> List[SuffixOnlyField]:
        """
        Returns all distinct suffix fields.

        :return:    The suffix fields.
        """
        # Get the unique prefix fields as a set
        result = {SuffixField(field) for field in self._params.keys() if field.is_compound()}

        # Put the prefix fields into a sorted list
        result = list(result)
        result.sort()

        return result

    def add_parameter(self, key: str, value: Any):
        """
        Add parameter value to store.

        :param key:     The key.
        :param value:   The value for the key.
        """
        # Get the string representation of this value
        value_string: str = str(value)

        # Get the field for this parameter
        field: Optional[AbstractField] = self._fields[key] if key in self._fields else None

        # Create a new field for the parameter if we don't have one already
        if field is None:
            fixed_string: str = Field.fix_string(value_string)
            datatype: DataType = DataType.UNKNOWN \
                if isinstance(value, str) \
                else DataType.guess_type(fixed_string)
            self._params[Field(key, datatype)] = fixed_string
        else:
            obj = field.value_of(value_string)
            if obj is None:
                return
            self._params[field] = obj if isinstance(value, str) else value

    def set_params(self, params: Dict[AbstractField, Any]):
        """
        Set the parameters.

        :param params:  Dict of parameters.
        """
        self._params = params

    def get_params(self) -> Dict[AbstractField, Any]:
        """
        Get the parameters.

        :return:    Dict of parameters.
        """
        return self._params

    def has_value(self, key: Union[AbstractField, str]) -> bool:
        """
        Returns whether a certain value is available in this report.

        :param key: The value to look for.
        :return:    True if the value is available.
        """
        if isinstance(key, str):
            key = Field(key, DataType.UNKNOWN)

        return self.get_value(key) is not None

    def has_values(self, fields: List[Union[AbstractField, str]]) -> bool:
        """
        Checks whether all the fields are available.

        :param fields:  The required fields.
        :return:        True if required fields are available.
        """
        return all(map(self.has_field, fields))

    def set_value(self, key: AbstractField, value: Any) -> bool:
        """
        Sets a value.

        :param key:     The key of the value.
        :param value:   The new value.
        :return:        Whether successfully set.
        """
        if isinstance(value, str):
            value = Field.fix_string(value)

        # Convert to the correct type
        try:
            value = key.datatype.convert(value)
        except Exception:
            return False

        self._params[key] = value
        if key.name in self._fields:
            self._fields[key.name] = key

        return True

    def set_numeric_value(self, key: str, value: Real):
        """
        Sets a numeric value.

        :param key:     The key of the value.
        :param value:   The new value.
        """
        self.set_value(Field(key, DataType.NUMERIC), value)

    def set_string_value(self, key: str, value: str):
        """
        Sets a string value.

        :param key:     The key of the value.
        :param value:   The new value.
        """
        self.set_value(Field(key, DataType.STRING), value)

    def set_boolean_value(self, key: str, value: bool):
        """
        Sets a boolean value.

        :param key:     The key of the value.
        :param value:   The new value.
        """
        self.set_value(Field(key, DataType.BOOLEAN), value)

    def get_value(self, key: AbstractField) -> Any:
        """
        Gets parameter value, or None if not available.

        :param key:     The key.
        :return:        Parameter value.
        """
        return self._params[key] if key in self._params else None

    def get_string_value(self, key: Union[str, AbstractField]) -> Optional[str]:
        """
        Gets parameter value, or None if not available.

        :param key:     The key.
        :return:        Parameter value.
        """
        if isinstance(key, str):
            key = Field(key, DataType.STRING)

        return str(self._params[key]) if key in self._params else None

    def get_boolean_value(self, key: Union[str, AbstractField]) -> Optional[bool]:
        """
        Gets parameter value, or None if not available.

        :param key:     The key.
        :return:        Parameter value.
        """
        if isinstance(key, str):
            key = Field(key, DataType.BOOLEAN)

        return self._params[key] if key in self._params and isinstance(self._params[key], bool) else None

    def get_real_value(self, key: Union[str, AbstractField]) -> Optional[Real]:
        """
        Gets parameter value, or None if not available.

        :param key:     The key.
        :return:        Parameter value.
        """
        if isinstance(key, str):
            key = Field(key, DataType.NUMERIC)

        return self._params[key] if key in self._params and isinstance(self._params[key], Real) else None

    def remove_value(self, key: AbstractField) -> Any:
        """
        Removes the specified field.

        :param key:     The key.
        :return:        The value previously stored in the report, or None if the field wasn't present.
        """
        return self._params.pop(key, None)

    def update(self):
        """
        Updates certain dependant fields. This method should be called before
        saving it to the database, after loading it from the database or when
        a quantitation report has been created by hand.

        Default implementation does nothing.
        """
        pass

    def to_string(self) -> str:
        """
        Returns the string representation of a report.

        :return:    String representation of report.
        """
        # Sort the fields into a list
        fields = list(self._params.keys())
        fields.sort()

        return "\n".join(
            f"{field.to_parseable_string()}: {self._params[field]}"
            for field in fields
        )

    def get_clone(self) -> "Report":
        """
        Returns a clone of itself.

        :return:    The clone.
        """
        # Create a new instance of this report type
        clone: Report = self.new_instance()

        # Assign our contents to the clone
        if clone is not None:
            clone.assign(self)

        return clone

    def assign(self, other: "Report"):
        """
        Obtains all the values from the specified report.

        :param other:   The report to obtain the values from.
        """
        self._fields = dict(other._fields)
        self._params = dict(other._params)
        self.database_id = other.database_id

    def set_dummy_report(self, value: bool):
        """
        Sets whether this report is dummy report or not.

        :param value:   If true then this report will be flagged as dummy report.
        """
        self.add_parameter(constants.FIELD_DUMMYREPORT, value)

    def is_dummy_report(self) -> bool:
        """
        Returns whether this report is a dummy report or not.

        :return:    True if this report is a dummy report.
        """
        return self.get_boolean_value(constants.FIELD_DUMMYREPORT) is True

    def compare_to(self, obj: Any) -> int:
        """
        Compares this object with the specified object for order.  Returns a
        negative integer, zero, or a positive integer as this object is less
        than, equal to, or greater than the specified object.

        :param obj:     The object to be compared.
        :return:        A negative integer, zero, or a positive integer as this object
                        is less than, equal to, or greater than the specified object.
        """
        raise NotImplementedError(Report.compare_to.__qualname__)

    def __eq__(self, other: Any):
        return isinstance(other, Report) and self._params == other._params

    def intersect(self, report: "Report") -> "Report":
        """
        Returns the intersection with this Quantitation Report and the provided
        one. The result contains the values of this quantitation report. No
        merging of values between the two is performed.

        :param report:  The report to get the intersection with.
        :return:        The intersection.
        """
        # Create a new report
        result: Report = self.new_instance()

        # Add the intersecting parameters
        if result is not None:
            result.set_params({key: self.get_value(key)
                               for key in self.get_fields()
                               if report.has_value(key)})

        return result

    def minus(self, report: "Report") -> "Report":
        """
        Returns the subset of values that this Quantitation Report contains, but
        not the provided one. The result contains the values of this quantitation
        report.

        :param report:  The report to get the fields to exclude from.
        :return:        The new report.
        """
        # Create a new report
        result: Report = self.new_instance()

        # Add the non-intersecting parameters
        if result is not None:
            result.set_params({key: self.get_value(key)
                               for key in self.get_fields()
                               if not report.has_value(key)})

        return result

    def merge_with(self, other: "Report"):
        """
        Merges its own data with the one provided by the specified object.
        Never overwrites its own values, only adds missing ones.

        :param other:   The object to merge with.
        """
        for field in other.get_fields():
            if not self.has_value(field):
                self.set_value(field, other.get_value(field))

    @classmethod
    def new_instance(cls) -> Optional["Report"]:
        """
        Returns a new instance of the concrete subclass for the report.

        :return:    The new instance.
        """
        try:
            return cls()
        except Exception:
            return None

    def to_properties(self) -> Dict[str, Any]:
        """
        Turns the report into a properties object. Also adds the parent ID.

        :return:
        """
        # Create properties with the parent ID
        result: Dict[str, Any] = {constants.PROPERTY_PARENTID: str(self.database_id)}

        # Transfer properties
        for field in self.get_fields():
            value = self.get_value(field)
            result[field.name] = str(value)
            result[field.name + constants.DATATYPE_SUFFIX] = str(DataType.guess_type(value))

        return result

    def __iter__(self):
        return iter(self.get_fields())

    def __contains__(self, item):
        return self.has_field(item)

    def __getitem__(self, item):
        return self.get_value(item)

    def __setitem__(self, key, value):
        self.set_value(key, value)

    def __str__(self):
        return self.to_string()



from typing import List, Union, Iterator, Dict, Iterable

from ._save_helper import quoted
from .._NamedColumnSelection import NamedColumnSelection, SELECTION_TYPE
from ._Attribute import Attribute, StringAttribute, NominalAttribute, NumericAttribute
from . import constants
from ... import onehot
from ...sequence import extract_by_index
from ...iterate import invert_indices
from ._error import ARFFError

# The type of the data section
VALUE_TYPE = Union[None, float, str]
ROW_TYPE = List[VALUE_TYPE]
DATA_TYPE = List[ROW_TYPE]

# Allow our selections to be raw or wrapped
SELECTION_TYPE = Union[NamedColumnSelection, SELECTION_TYPE]


class ARFFFile:
    """
    Represents an ARFF file in memory. Loads the file into memory on initialisation
    so it can be closed straight away.

    Currently does not handle relational attributes or date formats.
    """
    def __init__(self,
                 relation_name: str,
                 attributes: List[Attribute],
                 data: DATA_TYPE):
        self.relation_name: str = relation_name
        self._attributes: List[Attribute] = attributes
        self._data: DATA_TYPE = data

    def remove_row(self, row: int) -> ROW_TYPE:
        """
        Removes a single row from this ARFF file.

        :param row:     The index of the row to remove.
        :return:        The row itself.
        """
        return self._data.pop(row)

    def remove_rows(self, rows: Iterable[int]) -> DATA_TYPE:
        """
        Removes the given rows from this ARFF file and returns
        them. If indices are duplicated, they will be returned
        multiple times but only removed once.

        :param rows:    The rows to remove.
        :return:        The removed rows in the order given.
        """
        # Consume the iterable in case it's an iterator
        rows: List[int] = list(rows)

        # Save the result references first
        result: DATA_TYPE = extract_by_index(self._data, rows)

        # Construct the ordered unique list of rows to remove
        ordered_unique_rows = list(set(rows))
        ordered_unique_rows.sort()

        # Remove each row in reverse order (to not upset
        # ordering)
        for row in reversed(ordered_unique_rows):
            self.remove_row(row)

        return result

    def extract_rows(self, rows: Iterable[int]) -> 'ARFFFile':
        """
        Extracts a new ARFF file from this one consisting of the
        given rows in this file.

        :param rows:    The rows to extract (can contain duplicates).
        :return:        A new ARFF file with the given rows as its data.
        """
        return ARFFFile(
            self.relation_name,
            [attribute.copy() for attribute in self._attributes],
            [row.copy() for row in extract_by_index(self._data, rows)]
        )

    def extract_and_remove_rows(self, rows: Iterable[int]) -> 'ARFFFile':
        """
        Extracts a new ARFF file from this one consisting of the
        given rows in this file, and removes those rows from this
        ARFF file.

        :param rows:    The rows to extract (can contain duplicates).
        :return:        A new ARFF file with the given rows as its data.
        """
        return ARFFFile(
            self.relation_name,
            [attribute.copy() for attribute in self._attributes],
            self.remove_rows(rows)
        )

    def add_attribute(self, attribute: Attribute, at_index: int = -1):
        """
        Adds a new attribute to this ARFF file.

        :param attribute:   The attribute to add.
        :param at_index:    The index at which to insert the attribute.
        """
        # Make sure we don't already have an attribute with this name
        if attribute.name in self.attribute_names():
            raise ARFFError("Attribute '" + attribute.name + "' already exists in ARFF file")

        # Insert the attribute
        self._attributes.insert(at_index, attribute)

        # Insert missing values for the attribute in the data
        for row in self._data:
            row.insert(at_index, None)

    def remove_attributes(self, selection: SELECTION_TYPE):
        """
        Removes attributes and the associated data from the file.

        :param selection:   The selection of the attributes to remove.
        """
        # Get the indices of the attributes to remove
        remove_indices = self.selection_indices(selection)

        # Get the indices of the attributes to keep
        keep_indices = list(invert_indices(remove_indices, self.attribute_count()))

        # Remove the selected indices by extraction
        self._attributes = extract_by_index(self._attributes, keep_indices)

        # Remove the selected columns from the data
        for row_index in range(self.row_count()):
            self._data[row_index] = list(extract_by_index(self._data[row_index], keep_indices))

    def set_value(self, attribute_selection: SELECTION_TYPE, row: int, value: VALUE_TYPE):
        """
        Sets the value of the selected attribute in the given row.

        :param attribute_selection:     The selected attribute.
        :param row:                     The index of the row to set.
        :param value:                   The value to set.
        """
        # Get the index of the selected attribute
        attribute_index = self.selection_index(attribute_selection)

        # Get the attribute itself
        attribute: Attribute = self._attributes[attribute_index]

        # Parse the value to the correct type
        if value is not None:
            value = attribute(value)

        # Set the value in the data
        self._data[row][attribute_index] = value

    def get_value(self, attribute_selection: SELECTION_TYPE, row: int) -> VALUE_TYPE:
        """
        Gets the value of the given attribute in the given row.

        :param attribute_selection:     The attribute to get the value of.
        :param row:                     The row to get the value from.
        :return:                        The value.
        """
        # Get the index of the selected attribute
        attribute_index = self.selection_index(attribute_selection)

        return self._data[row][attribute_index]

    def row_count(self):
        """
        Gets the number of rows in the file.

        :return:    The number of rows.
        """
        return len(self._data)

    def attribute_count(self):
        """
        Gets the number of attributes in the file.

        :return:    The number of attributes.
        """
        return len(self._attributes)

    def map_string_attribute(self, selection: SELECTION_TYPE) -> List[str]:
        """
        Converts a string/nominal attribute to a numerical attribute where each entry is
        the index of the original string in a lookup table.

        :param selection:       The selection of the string/nominal attribute to map.
        :return:                The lookup table.
        """
        # Create the string index map for the selected attribute
        string_index_map = self.create_string_index_map(selection)

        # Get the selected attribute and its index
        attribute_index = self.selection_index(selection)
        attribute = self._attributes[attribute_index]

        # Create a table of strings from the index map (use the nominal ordering if a nominal attribute)
        string_table = attribute.ordered_values() if isinstance(attribute, NominalAttribute) else list(string_index_map)

        # Convert the string data to numeric
        for string_index, string in enumerate(string_table):
            numeric_value = float(string_index)
            row_indices = string_index_map[string]
            for row_index in row_indices:
                self._data[row_index][attribute_index] = numeric_value

        # Convert the attribute to a numeric attribute
        self._attributes[attribute_index] = NumericAttribute(attribute.name)

        # Return the lookup table
        return string_table

    def create_string_index_map(self, selection: SELECTION_TYPE) -> Dict[str, List[int]]:
        """
        Creates a mapping from string values to row indices for a string attribute
        or a nominal attribute.

        :param selection:   The selection of attribute to map.
        :return:            The index map for the strings in the given attribute.
        """
        # Get the index of the selected attribute
        attribute_index = self.selection_index(selection)

        # Get the attribute itself
        attribute = self._attributes[attribute_index]

        # Make sure the selected attribute is a string or nominal attribute
        if not (isinstance(attribute, StringAttribute) or isinstance(attribute, NominalAttribute)):
            raise ValueError("Can only map string or nominal attributes")

        # Create the map
        string_index_map = {}

        # Add each value to the map
        for row_index, row in enumerate(self._data):
            string_value = row[attribute_index]

            if string_value not in string_index_map:
                string_index_map[string_value] = []

            string_index_map[string_value].append(row_index)

        # Return the map
        return string_index_map

    def one_hot_encode(self):
        """
        Applies a one-hot encoding to all nominal attributes in the file.

        :return:    The mapping used to perform the encoding.
        """
        # Create the mapping from nominal to numeric attributes
        one_hot_map = create_one_hot_map(self._attributes)

        # Apply the mapping to the attributes
        self._attributes = one_hot_encode_attributes(self._attributes, one_hot_map)

        # Apply the mapping to the data
        one_hot_map.encode(self._data)

        # Return the one-hot encoding
        return one_hot_map

    def extract_data(self, selection: SELECTION_TYPE) -> DATA_TYPE:
        """
        Extracts the data for a selection of attributes/columns.

        :param selection:   The selection of attributes/columns.
        :return:            The data.
        """
        # Extract the selected columns from the data
        selection = self.selection_indices(selection)
        return [extract_by_index(row, selection) for row in self._data]

    def extract_all_data(self) -> DATA_TYPE:
        """
        Extracts a copy of this ARFF file's data.
        """
        return [row.copy() for row in self._data]

    def selection_attributes(self, selection: SELECTION_TYPE) -> List[Attribute]:
        """
        Gets the attributes selected by the given selection.

        :param selection:   The selection.
        :return:            The attributes.
        """
        return extract_by_index(self._attributes, self.selection_indices(selection))

    def selection_indices(self, selection: SELECTION_TYPE) -> List[int]:
        """
        Gets the attribute indices selected by the given selection.

        :param selection:   The selection.
        :return:            The attribute indices.
        """
        # Wrap any raw selection types
        if not isinstance(selection, NamedColumnSelection):
            selection = NamedColumnSelection(selection)

        return selection.apply(list(self.attribute_names()))

    def selection_index(self, selection: SELECTION_TYPE) -> int:
        """
        Gets the index of the attribute selected by the given selection.
        Enforces that one and only one attribute is selected.

        :param selection:   The selection.
        :return:            The attribute index.
        """
        # Get the selected indices
        attribute_indices = self.selection_indices(selection)

        # Make sure only one is selected
        if len(attribute_indices) != 1:
            raise ValueError("Must select one and only one attribute")

        # Return the single selected index
        return attribute_indices[0]

    def attribute_names(self) -> Iterator[str]:
        """
        Iterates over the names of the attributes.
        """
        return (attribute.name for attribute in self._attributes)

    def relation_section(self) -> str:
        """
        Gets the relation section of this ARFF file as a string.
        """
        return constants.RELATION_SECTION_KEYWORD + " " + quoted(self.relation_name)

    def attributes_section(self) -> str:
        """
        Gets the attributes section of this ARFF file as a string.
        """
        return "\n".join(attribute.to_string() for attribute in self._attributes)

    def data_section(self) -> str:
        """
        Gets the data section of this ARFF file as a string.
        """
        return constants.DATA_SECTION_KEYWORD + "\n" + \
            "\n".join(
                ",".join(
                    quoted(str(value)) if value is not None else constants.MISSING_VALUE_SYMBOL
                    for value in row
                )
                for row in self._data
            )

    def __str__(self) -> str:
        return self.relation_section() + "\n\n" + \
               self.attributes_section() + "\n\n" + \
               self.data_section() + "\n"


def create_one_hot_map(attributes: List[Attribute]) -> onehot.Mapping:
    """
    Creates a list of one-hot encoding maps for the given attributes.

    :param attributes:      The attributes being encoded.
    :return:                The list of one-hot mappings.
    """

    # Create the map
    one_hot_map = onehot.Mapping(len(attributes))

    # Process each attribute
    for attribute_index, attribute in enumerate(attributes):
        # If the attribute is nominal, add a mapping from value name to encoding
        if isinstance(attribute, NominalAttribute):
            values = attribute.value_iterator()

            # Append the mapping structure to the one-hot map
            one_hot_map.add_encoding(attribute_index, onehot.Encoding(values, float))

    # Return the map
    return one_hot_map


def one_hot_encode_attributes(attributes: List[Attribute], one_hot_map: onehot.Mapping) -> List[Attribute]:
    """
    Uses the one-hot encoding map to modify the given attribute list into its
    one-hot encoded form.

    :param attributes:      The attributes to encode.
    :param one_hot_map:     The one-hot encoding map to use for the encoding.
    :return:                The encoded list of attributes.
    """

    attribute_names = [attribute.name for attribute in attributes]
    attribute_lookup = {attribute.name: attribute for attribute in attributes}

    new_names = one_hot_map.encode_header(attribute_names)

    for name in new_names:
        if name not in attribute_lookup:
            attribute_lookup[name] = NumericAttribute(name)

    return [attribute_lookup[name] for name in new_names]



"""
Module for constants relating to locating objects.
"""
# The key for the X location
KEY_X: str = "x"

# The key for the Y location
KEY_Y: str = "y"

# The key for the width
KEY_WIDTH: str = "width"

# The key for the height
KEY_HEIGHT: str = "height"

# The key for the location
KEY_LOCATION: str = "location"

# The key for the Xs of the polygon in the meta-data (comma-separated list)
KEY_POLY_X: str = "poly_x"

# The key for the Ys of the polygon in the meta-data (comma-separated list)
KEY_POLY_Y: str = "poly_y"

# The key for the overall count
KEY_COUNT: str = "count"

# The key for the index of a group
KEY_INDEX: str = "index"



def factorial(
        of: int,
        down_to: int = 0
) -> int:
    """
    Returns the multiplication of all positive integers from ``of`` down
    to (but not including) ``down_to``.

    :param of:
                The greatest positive integer to include in the product.
    :param down_to:
                The greatest positive integer, less than ``of``, to exclude
                from the product.
    :return:
                The factorial of ``of`` down to ``down_to``. If ``of`` equals
                ``down_to``, the result is ``of``.
    """
    if down_to < 0:
        raise ArithmeticError(f"'down_to' cannot be less than 0, got {down_to}")
    if of < down_to:
        raise ArithmeticError(f"'of' must be at least 'down_to', got 'of = {of}, 'down_to' = {down_to}")

    result = 1
    while of > down_to:
        result *= of
        of -= 1

    return result



from typing import Dict, Set, Type, Optional, Iterator

from .meta import fully_qualified_name
from .meta.code_repr import (
    CodeRepresentable,
    CodeRepresentation,
    code_repr,
    get_import_dict,
    get_code,
    combine_import_dicts
)


class ClassRegistry(CodeRepresentable):
    """
    A registry of classes that can be looked up by name.

    # TODO: Alias conflicts with name/qualname
    """
    def __init__(self, *classes: Type):
        self._name_lookup: Dict[str, Set[str]] = {}
        self._qualname_lookup: Dict[str, Type] = {}
        self._aliases: Dict[str, str] = {}
        self._alias_reverse_lookup: Dict[str, str] = {}

        # Register any initial classes
        for cls in classes:
            self.register(cls)

    def register(self, cls: Type):
        """
        Registers a class with the registry.

        :param cls:     The class to register.
        :return:        The fully-qualified name of the class.
        """
        # Get the short and qualified names of the class
        name: str = cls.__name__
        qualname: str = fully_qualified_name(cls)

        # Abort if already registered
        if qualname in self._qualname_lookup:
            return qualname

        # Initialise the short-name list of necessary
        if name not in self._name_lookup:
            self._name_lookup[name] = set()

        # Add the qualified name to the lookup
        self._name_lookup[name].add(qualname)

        # Add the type to the qualified lookup
        self._qualname_lookup[qualname] = cls

        return qualname

    def is_registered(self, cls: Type) -> bool:
        """
        Checks if the given class is registered.

        :param cls:     The class to check.
        :return:        True if the class is registered,
                        False if not.
        """
        return fully_qualified_name(cls) in self._qualname_lookup

    def registered_classes(self) -> Iterator[Type]:
        """
        Gets an iterator over all registered classes.
        """
        return (value for value in self._qualname_lookup.values())

    def alias(self, cls: Type, alias: str) -> 'ClassRegistry':
        """
        Creates an alias for a class. Automatically registers
        the class if it's not already. Chainable.

        :param cls:         The class to alias.
        :param alias:       The alias to use.
        :raises NameError:  If the alias is already in use, or is empty.
        """
        # Make sure the alias isn't the empty string
        if alias == "":
            raise NameError("Aliases can't be empty")

        # Make sure the alias isn't already in use
        if alias in self._aliases:
            raise NameError(f"Alias '{alias}' already in use")

        # Make sure the class is registered
        qualname = self.register(cls)

        # Save the alias
        self._aliases[alias] = qualname
        self._alias_reverse_lookup[qualname] = alias

        # Return self so as to be chainable
        return self

    def resolve(self, name: str) -> Optional[str]:
        """
        Attempts to resolve names to their fully-qualified equivalent.

        :param name:        The name to resolve (qualified/short/alias).
        :return:            The fully-qualified name, or None if not found.
        :raises NameError:  If the name is an ambiguous short-name.
        """
        # If the name is an alias, dereference it
        if name in self._aliases:
            return self._aliases[name]

        # If the name is qualified, return the class
        if name in self._qualname_lookup:
            return name

        # If the name is a short name, try to resolve it
        if name in self._name_lookup:
            # Get the potential qualified names
            qualnames = self._name_lookup[name]

            # If there's more than one, it's ambiguous
            if len(qualnames) > 1:
                raise NameError(f"Class name '{name}' is ambiguous: {', '.join(qualnames)}")

            for qualname in qualnames:  # Dummy loop to get first (only) item from set
                return qualname

        return None

    def find(self, name: str, *,
             attempt_class_import: bool = False) -> Optional[Type]:
        """
        Tries to find the registered class with the given name.

        :param name:                    The name to find (short or qualified).
        :param attempt_class_import:    Whether to attempt to import the class
                                        if it's not registered.
        :return:                        The class with the given name, or None
                                        if not found.
        :raises NameError:              If the name is an ambiguous short-name.
        """
        # Resolve the name to its fully-qualified counterpart
        qualname = self.resolve(name)

        # If resolution succeeded, return the named class
        if qualname is not None:
            return self._qualname_lookup[qualname]

        # Attempt to import the class if selected
        if attempt_class_import:
            return self.attempt_class_import(name)

        # Class not found
        return None

    def aliases(self) -> Iterator[str]:
        """
        Gets an iterator over the aliases registered with the registry.

        :return:    The iterator of aliases.
        """
        return (str(name) for name in self._aliases.keys())

    def has_alias(self, alias: str) -> bool:
        """
        Whether the registry has an alias with the given name.

        :param alias:   The alias to check for.
        :return:        True if the alias exists, False if not.
        """
        return alias in self._aliases

    def get_alias(self, cls: Type) -> Optional[str]:
        """
        Gets the alias for a class.

        :param cls:     The class to get the alias for.
        :return:        The alias, or None if the class is not aliased.
        """
        # Get the class' fully-qualified name
        qualname = fully_qualified_name(cls)

        return self._alias_reverse_lookup.get(qualname, None)

    def attempt_class_import(self, name: str) -> Optional[Type]:
        """
        Attempts to import the class by name.

        :param name:    The fully-qualified name of the class.
        :return:        The class, or None if it could not
                        be imported.
        """
        # TODO: Implementation
        raise NotImplementedError(ClassRegistry.attempt_class_import.__qualname__)

    def code_repr(self) -> CodeRepresentation:
        # Get the code-representations of all the types that are in the registry
        aliased_types = {}
        unaliased_types = []
        for cls in self._qualname_lookup.values():
            alias = self.get_alias(cls)
            if alias is not None:
                aliased_types[get_code(code_repr(alias))] = code_repr(cls)
            else:
                unaliased_types.append(code_repr(cls))

        # Combine the import dictionaries of the ClassRegistry class and all registered classes
        import_dict = combine_import_dicts(get_import_dict(code_repr(type(self))),
                                           *map(get_import_dict, unaliased_types),
                                           *map(get_import_dict, aliased_types.values()))

        # Create the code representation for just unaliased types (using constructor)
        code = f"ClassRegistry({', '.join(map(get_code, unaliased_types))})"

        # Add the chained calls to 'alias' to add the aliased types
        for alias, code_representation in aliased_types.items():
            code += f".alias({get_code(code_representation)}, {alias})"

        return import_dict, code



from typing import IO

from .._FileWriter import FileWriter
from ._CSVFile import CSVFile


class CSVFileWriter(FileWriter[str, CSVFile, "CSVFileWriter"]):
    """
    Writer for CSV files.
    """
    def _dump(self, obj: CSVFile, file: IO[str]):
        file.write(str(obj))



from ._NamedColumnSelection import NamedColumnSelection, SELECTION_TYPE
from ._functions import load_dir
from ._PathContextManager import PathContextManager, offset_cwd
from ._util import ensure_path



from ._get_argument_to_typevar import get_argument_to_typevar
from ._TypeVarProperty import TypeVarProperty
from ._typing import (
    POSITIONAL_ARGS_TYPE,
    KEYWORD_ARGS_TYPE,
    VAR_ARGS_TYPE,
    AnyCallable,
    ArgType,
    ReturnType,
    GenericCallable,
    GenericDecorator
)



from abc import ABC, abstractmethod
from argparse import Action, ArgumentParser
from typing import Type, Union, Any, Iterable, Tuple, Dict

from ...meta import non_default_kwargs
from ...meta.code_repr import CodeRepresentable, CodeRepresentation
from ..util import is_flag_reason, is_short_flag, is_long_flag, flag_from_name
from .._ArgumentParserConfigurer import ArgumentParserConfigurer
from .._typing import OptionsList


class Option(CodeRepresentable, ArgumentParserConfigurer, ABC):
    """
    Descriptor class which sets an option on an option-handler.
    """
    def __init__(self,
                 code_representation: CodeRepresentation,
                 *flags: str,
                 action: Union[str, Type[Action]] = ...,
                 nargs: Union[int, str] = ...,
                 default: Any = ...,
                 choices: Iterable[Any] = ...,
                 required: bool = ...,
                 metavar: Union[str, Tuple[str, ...]] = ...,
                 help: str = ...):
        # Check the flags are supplied correctly
        self._check_flags(flags)

        self._name: str = ""
        self._owner: Type = type(None)
        self._flags: Tuple[str] = flags

        # Save the optional flag
        self._optional: bool = not required if required is not ... else True

        # Convert the default value if supplied
        if default is not ...:
            # Can't set a default value for a required option
            if not self._optional:
                raise ValueError(f"Can't set a default value for a required option")
        self._default = default

        # Save any non-default arguments as kwargs
        self._kwargs: Dict[str, Any] = non_default_kwargs(Option.__init__, locals())

        # The code representation of the option
        self._code_representation: CodeRepresentation = code_representation

    # =========== #
    # NAMES/FLAGS #
    # =========== #

    @property
    def name(self) -> str:
        """
        Gets the name of this option.

        :return: The option's name.
        """
        return self._name

    def _require_name(self, message: str):
        """
        Raises an exception with the provided message if the name
        for this option has not yet been set.

        :param message:     The message to raise.
        :raises NameError:  If the name has not been set yet.
        """
        if self._name == "":
            raise NameError(message)

    @property
    def flags(self) -> Tuple[str]:
        """
        Gets the flags that correspond to this option.

        :return:    The option string.
        """
        return self._flags

    def _require_flags(self, message: str):
        """
        Raises an exception with the provided message if the flags
        for this option has not yet been set.

        :param message:     The message to raise.
        :raises NameError:  If the flags have not been set yet.
        """
        if len(self._flags) == 0:
            raise NameError(message)

    @staticmethod
    def _check_flags(flags: Tuple[str]):
        """
        Makes sure the flags given to __init__ are correct.

        :param flags:   The flags as given to __init__.
        """
        # Get the number of flags given
        num_flags = len(flags)

        # If no flags are given, a long flag will be inferred from the attribute name
        if num_flags == 0:
            return

        # Can only have 2 flags at most (1 short and 1 long)
        if num_flags > 2:
            raise ValueError("Provided too many flags (can only have 2)")

        # Make sure the flags are strings
        for flag in flags:
            if not isinstance(flag, str):
                raise TypeError(f"Flags must be strings but got a {type(flag).__qualname__}")

        # Make sure the flags are in fact flags
        for flag in flags:
            flag_reason = is_flag_reason(flag)
            if flag_reason is not None:
                raise ValueError(f"'{flag}' is not a flag: {flag_reason}")

        # If two flags given, make sure one is short and one is long
        if num_flags == 2:
            if (
                    (is_short_flag(flags[0]) and is_short_flag(flags[1])) or
                    (is_long_flag(flags[0]) and is_long_flag(flags[1]))
            ):
                raise ValueError(f"If providing 2 flags, one must be a short flag and the other a long flag")

    def __set_name__(self, owner, name: str):
        # Only optional handlers can use options
        from .._OptionHandler import OptionHandler
        if not issubclass(owner, OptionHandler):
            raise TypeError(f"Class {owner.__name__} is not an OptionHandler")

        # Can't have private options
        if name.startswith("_"):
            raise ValueError(f"Option names can't start with underscores")

        # Can't reuse options (use a copy instead)
        if self._name != "":
            # If assigning the same name (e.g. in generic base class), just ignore
            if self._name == name:
                return

            raise NameError(f"Can't reuse options; use a copy instead")

        # If no flags were specified, create one from the name
        if len(self._flags) == 0:
            self._flags = (flag_from_name(name),)

        # Save our name
        self._name = name
        self._owner = owner

    # ========== #
    # PROPERTIES #
    # ========== #

    @property
    def optional(self) -> bool:
        """
        Whether this option is optional.
        """
        return self._optional

    @property
    def default(self) -> Any:
        """
        Gets the default value for this option.

        :return:    The default value.
        """
        # Can't get the default value for a required option
        if not self._optional:
            raise RuntimeError("Can't get the default value of a required option")

        # Parse the default using argparse if none was set
        if self._default is ...:
            self._default = self._get_namespace_value_from_options_list([])

        return self._default

    @property
    def kwargs(self) -> Dict[str, Any]:
        """
        Gets the kwargs for this option.

        :return:    The kwargs.
        """
        # Take a copy so that the kwargs can't be mutated
        return dict(self._kwargs)

    @property
    def owner(self) -> Type:
        """
        Gets the class that owns this option.

        :return: A sub-type of OptionHandler.
        """
        # Can't get the owner until after binding
        self._require_name("Can't get the owner of an unbound option")

        return self._owner

    # =================== #
    # CODE REPRESENTATION #
    # =================== #

    def code_repr(self) -> CodeRepresentation:
        return self._code_representation

    # ====== #
    # VALUES #
    # ====== #

    def __get__(self, instance, owner):
        # If called on the class, return the option itself
        if instance is None:
            return self

        # Can't get the value until the option has been bound
        self._require_name("Can't get option value until it is bound")

        # If called on an instance, the instance must be an option-value handler
        from .._OptionValueHandler import OptionValueHandler
        if not isinstance(instance, OptionValueHandler):
            raise TypeError(f"Instance '{instance}' is not an option-value handler")

        # Get the option's value from the namespace
        namespace_value = getattr(instance.namespace, self.name)

        # If it is the default value, just return it
        if self._optional and namespace_value is self.default:
            return namespace_value

        # Validate the namespace value
        self._validate_namespace_value(namespace_value)

        return self._namespace_value_to_internal_value(namespace_value)

    @abstractmethod
    def _namespace_value_to_internal_value(self, namespace_value: Any) -> Any:
        """
        Converts the namespace value for this option into its internal type.

        :param namespace_value:     The option's value from the namespace.
        :return:                    The option's internal value.
        :raises Exception:          If the namespace value is invalid.
        """
        pass

    def __set__(self, instance, value):
        # Must have a name set to access the namespace
        self._require_name("Can't set the value until the option is bound")

        # The instance must be an option-value handler
        from .._OptionValueHandler import OptionValueHandler
        if not isinstance(instance, OptionValueHandler):
            raise TypeError(f"Instance '{instance}' is not an option-value handler")

        # Make sure the value is valid
        self._validate_internal_value(value)

        # Convert to a namespace value
        namespace_value = self._internal_value_to_namespace_value(value)

        # Set the option's value
        setattr(instance.namespace, self.name, namespace_value)

    @abstractmethod
    def _validate_internal_value(self, internal_value: Any):
        """
        Checks that the internal value is valid.

        :param internal_value:  The value to check.
        """
        pass

    @abstractmethod
    def _internal_value_to_namespace_value(self, internal_value: Any) -> Any:
        """
        Converts an internal value for this option into a namespace
        value.

        :param internal_value:  The value attempting to be set.
        :return:                The actual value that should be set.
        :raises Exception:      If the value cannot be validated.
        """
        pass

    def get_as_options_list(self, instance) -> OptionsList:
        """
        Gets the value of this option as an options list.

        :param instance:    The option value handler to get the value from.
        :return:            The options-list representation.
        """
        # Make sure we have flags to use
        self._require_name("Can't get option value as option list before the option has been bound")

        # The instance must be an option-value handler
        from .._OptionValueHandler import OptionValueHandler
        if not isinstance(instance, OptionValueHandler):
            raise TypeError(f"Instance '{instance}' is not an option-value handler")

        # Get the namespace value of the option
        namespace_value = getattr(instance.namespace, self.name)

        # If it's the default value, return an empty options-list
        if self._optional and namespace_value is self.default:
            return []

        # Validate the namespace value
        self._validate_namespace_value(namespace_value)

        # Let the sub-type format the options-list from the value
        return self._namespace_value_to_options_list(namespace_value)

    @abstractmethod
    def _validate_namespace_value(self, namespace_value: Any):
        """
        Checks the namespace value is valid.

        :param namespace_value: The namespace value to check.
        """
        pass

    @abstractmethod
    def _namespace_value_to_options_list(self, namespace_value: Any) -> OptionsList:
        """
        Converts the namespace value of this option into an options list.

        :param namespace_value:     The namespace value of the option.
        :return:                    The options-list representation of the value.
        """
        pass

    def set_from_options_list(self, instance, options_list: OptionsList):
        """
        Sets the value of this option from an options-list.

        :param instance:        The option value handler to set the value against.
        :param options_list:    The options-list value to set.
        """
        # Must have a name set to access the namespace
        self._require_name("Can't set the value via options-list until the option is bound")

        # The instance must be an option-value handler
        from .._OptionValueHandler import OptionValueHandler
        if not isinstance(instance, OptionValueHandler):
            raise TypeError(f"Instance '{instance}' is not an option-value handler")

        # Set it on the instance's namespace
        setattr(instance.namespace, self.name, self._get_namespace_value_from_options_list(options_list))

    def _get_namespace_value_from_options_list(self, options_list: OptionsList):
        """
        Gets the namespace value parsed from the given options-list for
        this option.

        :param options_list:    The options list.
        :return:                The namespace value.
        """
        # Must have a name set to access the namespace
        self._require_name("Can't get a namespace value via options-list until the option is bound")

        # Get a parser to parse the options list
        parser = self.get_configured_parser(add_help=False)

        # Parse the options list
        try:
            namespace = parser.parse_args(options_list)
        except SystemExit as e:
            raise ValueError("Error parsing options list") from e

        return getattr(namespace, self.name)

    # ===== #
    # OTHER #
    # ===== #

    def __eq__(self, other) -> bool:
        # Must be an option
        if not isinstance(other, Option):
            return False

        return self is other or self._kwargs == other._kwargs

    def configure_parser(self, parser: ArgumentParser):
        # Need a name to configure the parser's 'dest' argument
        self._require_name("Cannot configure a parser before the option is bound")

        parser.add_argument(*self._flags, dest=self._name, **self._kwargs)



from typing import Callable, Any, List


class LazyDescriptor:
    """
    Descriptor which instantiates another descriptor only when it is
    first accessed.
    """
    def __init__(self, descriptor_constructor: Callable[[], Any]):
        # The function that will create the descriptor when needed
        self._descriptor_constructor: Callable[[], Any] = descriptor_constructor

        # Any time we might have been attached to something
        self._names: List[str] = []
        self._owners: List = []

    def __get__(self, instance, owner):
        # Resolve and defer
        return self._resolve().__get__(instance, owner)

    def __set__(self, instance, value):
        # Resolve and defer
        self._resolve().__set__(instance, value)

    def __delete__(self, instance):
        # Resolve and defer
        self._resolve().__delete__(instance)

    def __set_name__(self, owner, name):
        # Save the arguments for actual construction time
        self._names.append(name)
        self._owners.append(owner)

    def _resolve(self):
        # Create the actual descriptor
        descriptor = self._descriptor_constructor()

        # If we're attached to anything, replace ourselves with the new descriptor
        for name, owner in zip(self._names, self._owners):
            if getattr(owner, name, None) is self:
                setattr(owner, name, descriptor)

        return descriptor



from typing import List, Generic, TypeVar, IO

from .._Serialiser import Serialiser
from ._IntSerialiser import IntSerialiser

ElementType = TypeVar("ElementType")


class ListSerialiser(
    Serialiser[List[ElementType]],
    Generic[ElementType]
):
    def __init__(
            self,
            element_serialiser: Serialiser[ElementType],
            length_serialiser: Serialiser[int] = IntSerialiser(signed=False)
    ):
        super().__init__()

        self._element_serialiser = element_serialiser
        self._length_serialiser = length_serialiser

    def _check(
            self,
            obj: List[ElementType]
    ):
        # Must be a list
        if not isinstance(obj, list):
            raise ValueError(f"{ListSerialiser.__class__.__name__} serialises lists")

    def _serialise(
            self,
            obj: List[ElementType],
            stream: IO[bytes]
    ):
        # Serialise the length of the list
        self._length_serialiser.serialise(len(obj), stream)

        # Serialise each value in turn
        for value in obj:
            self._element_serialiser.serialise(value, stream)

    def _deserialise(
            self,
            stream: IO[bytes]
    ) -> List[ElementType]:
        # Deserialise the length of the list
        length = self._length_serialiser.deserialise(stream)

        # Deserialise 'length' number of elements in serialised order
        return [
            self._element_serialiser.deserialise(stream)
            for _index in range(length)
        ]



import os
from abc import abstractmethod
from io import BytesIO
from typing import Generic, TypeVar, IO

# The type of object the serialiser serialises/deserialises
ObjectType = TypeVar("ObjectType")


class Serialiser(Generic[ObjectType]):
    """
    Base class for objects which serialise other objects
    to/from a binary representation.
    """
    # ============= #
    # Serialisation #
    # ============= #
    @abstractmethod
    def _check(self, obj: ObjectType):
        """
        Checks the object is suitable for serialisation.

        :param obj:             The object to check.
        :raises Exception:      If the object is not suitable for serialisation.
        """
        pass

    @abstractmethod
    def _serialise(self, obj: ObjectType, stream: IO[bytes]):
        """
        Serialises the given object to the given data-stream.

        :param obj:         The object to serialise.
        :param stream:      The stream to write the object to.
        """
        pass

    def serialise(self, obj: ObjectType, stream: IO[bytes]):
        """
        Serialises the given object to the given data-stream.

        :param obj:         The object to serialise.
        :param stream:      The stream to write the object to.
        """
        self._check(obj)
        self._serialise(obj, stream)

    def serialise_to_file(self, obj: ObjectType, filename: str):
        """
        Serialises the object to a file.

        :param obj:         The object to serialise.
        :param filename:    The name of the file to serialise to.
        """
        # Make sure the directory for the file exists
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        # Save the file
        with open(filename, 'wb') as file:
            self.serialise(obj, file)

    def serialise_to_bytes(self, obj: ObjectType) -> bytes:
        """
        Gets the serialised state of the object.

        :param obj:     The object to serialise.
        """
        # Create an in-memory stream
        stream = BytesIO()

        # Serialise the object to the stream
        self.serialise(obj, stream)

        # Reset the stream
        stream.seek(0)

        # Return the stream contents
        return stream.read()

    # =============== #
    # Deserialisation #
    # =============== #

    @abstractmethod
    def _deserialise(self, stream: IO[bytes]) -> ObjectType:
        """
        Deserialises an object from the given data-stream.

        :param stream:  The stream to read the object from.
        :return:        The deserialised object.
        """
        pass

    def deserialise(self, stream: IO[bytes]) -> ObjectType:
        """
        Deserialises an object from the given data-stream.

        :param stream:  The stream to read the object from.
        :return:        The deserialised object.
        """
        return self._deserialise(stream)

    def deserialise_from_file(self, filename: str) -> ObjectType:
        """
        Deserialises an object from the given file.

        :param filename:    The filename.
        :return:            The deserialised object.
        """
        with open(filename, 'rb') as file:
            return self.deserialise(file)

    def deserialise_from_bytes(self, data: bytes) -> ObjectType:
        """
        Deserialises an object from its serialised state.

        :param data:    The serialised object.
        :return:        The deserialised object.
        """
        return self.deserialise(BytesIO(data))



from argparse import ArgumentParser
from sys import stderr
from typing import Optional, IO

from ...logging import LoggingMixin


class LoggingArgumentParser(ArgumentParser, LoggingMixin):
    """
    An argument parser that logs messages instead of printing
    them to stdout/stderr.
    """
    def _print_message(self, message: str, file: Optional[IO[str]] = None):
        if message:
            # Copy the base argument parser's default for file
            if file is None:
                file = stderr

            # Choose a logger based on the output stream
            if file is stderr:
                logger = self.logger.error
            else:
                logger = self.logger.warning

            # Log the message
            logger(message)



"""
Package for meta-programming constructs.
"""
from ._fully_qualified_name import fully_qualified_name
from ._functions import (
    has_been_overridden,
    get_class_from_function,
    is_class_function,
    unbind,
    does_not_raise,
    all_as_kwargs
)
from ._import_code import get_import_code
from ._instanceoptionalmethod import instanceoptionalmethod
from ._InstanceProperty import InstanceProperty
from ._metadata import with_metadata, get_metadata, has_metadata
from ._non_default_kwargs import non_default_kwargs



from abc import ABC, abstractmethod
from typing import IO

from .._Serialiser import Serialiser

# The null terminator byte
NULL_BYTE: bytes = b'\x00'


class StringSerialiser(Serialiser[str], ABC):
    """
    Base class for string serialisers.
    """
    def __init__(self, encoding: str = "utf-8"):
        self._encoding: str = encoding

    def _check(self, obj: str):
        # Make sure it really is a string
        if not isinstance(obj, str):
            raise TypeError(f"StringSerialiser serialises strings, got {type(obj)}")

    @abstractmethod
    def _check_encoded(self, encoded: bytes):
        """
        Checks the encoded string is suitable for this serialiser.

        :param encoded:     The encoded string.
        """
        pass

    def _serialise(self, obj: str, stream: IO[bytes]):
        # Encode the string
        encoded: bytes = obj.encode(self._encoding)

        # Check the encoded string
        self._check_encoded(encoded)

        # Serialise the encoded string
        self._serialise_encoded(encoded, stream)

    @abstractmethod
    def _serialise_encoded(self, encoded: bytes, stream: IO[bytes]):
        """
        Serialises the encoded string to the byte-stream.

        :param encoded:     The encoded string.
        :param stream:      The stream.
        """
        pass

    def _deserialise(self, stream: IO[bytes]) -> str:
        return self._deserialise_encoded(stream).decode(self._encoding)

    @abstractmethod
    def _deserialise_encoded(self, stream: IO[bytes]) -> bytes:
        """
        Deserialises the encoded string from the stream.

        :param stream:  The stream.
        :return:        The encoded string.
        """
        pass



from argparse import ArgumentParser
from typing import Type

from ..options import Option
from .._ArgumentParserConfigurer import ArgumentParserConfigurer
from .._OptionHandler import OptionHandler
from ._OptionContext import OptionContext


class OptionClassContext(OptionContext, ArgumentParserConfigurer):
    """
    Class which handles accessing an option in the context of
    a specific class.
    """
    def __init__(self, option: Option, context: Type[OptionHandler]):
        # The option must be owned by the context type (or super-type)
        if not issubclass(context, option.owner):
            raise TypeError("Context type does not have access to this option")

        super().__init__(option, context)

    def configure_parser(self, parser: ArgumentParser):
        # Get the kwargs from the option
        kwargs = dict(self._option.kwargs)

        # Get our help text from the owner if we don't have it already
        if "help" not in kwargs:
            help = self._context.get_help_text_for_option(self._option)
            if help is not None:
                kwargs["help"] = help

        parser.add_argument(*self._option.flags, dest=self._option.name, **kwargs)



from collections import UserList
from numbers import Real
from typing import List, Optional, Iterable, Union, Set, Iterator, Dict

from ....file.report import Report, DataType, Field, AbstractField
from ._LocatedObject import LocatedObject
from . import constants


class LocatedObjects(UserList):
    def __init__(self, objects: Optional[Iterable[LocatedObject]] = None):
        super().__init__(objects)

    def subset(self, indices: List[int]) -> "LocatedObjects":
        """
        Returns a new instance using the specified object indices.

        :param indices: The indices for the subset.
        :return:        The subset.
        """
        # Ensure unique indices
        indices = set(indices)

        return LocatedObjects((obj for obj in self if obj.get_index() in indices))

    def remove(self, item: Union[LocatedObject, List[int]]) -> None:
        """
        Removes the objects with the specified indices.

        :param item:    The indices to remove.
        """
        # List implementation
        if isinstance(item, LocatedObject):
            return super().remove(item)

        # Ensure unique indices
        indices: Set[int] = set(item)

        # Pop any objects with indices in the set
        i: int = 0
        while i < len(self):
            if self[i].get_index() in indices:
                self.pop(i)
            else:
                i += 1

    def find(self, index: Union[int, str]) -> Optional[LocatedObject]:
        """
        Returns the object with the specified index.

        :param index:   The index to look for.
        :return:        The object, None if not found.
        """
        # Get both a string and integer representation
        if isinstance(index, str):
            try:
                int_index = int(index)
            except Exception:
                int_index = None
        else:
            int_index = index
            index = str(index)

        for obj in self:
            # Try for exact match
            obj_index_string: str = obj.get_index_string()
            if obj_index_string is not None and obj_index_string == index:
                return obj

            # Try for numeric match
            if int_index is not None and obj.get_index() == int_index:
                return obj

        return None

    def scale(self, scale: float):
        """
        Scales all objects with the provided scale factor.

        :param scale:   The scale factor.
        """
        for obj in self:
            obj.scale(scale)

    def to_report(self, prefix: str = "Object", offset: int = 0, update_index: bool = False) -> Report:
        """
        Turns the located objects into a report. Using a prefix like "Object." will
        result in the following report entries for a single object:

        Object.1.x
        Object.1.y
        Object.1.width
        Object.1.height
        Object.1.poly_x -- if polygon data present
        Object.1.poly_y -- if polygon data present

        :param prefix:          The prefix to use.
        :param offset:          The offset for the index to use.
        :param update_index:    Whether to update the index in the metadata.
        :return:                The generated report.
        """
        # Make sure the prefix doesn't end in a dot
        if prefix.endswith("."):
            prefix = prefix[:-1]

        # Create the empty report
        result: Report = Report()

        # Create a shortcut for adding values
        def add_value(count_string: str, suffix: str, type: DataType, value):
            field: Field = Field(f"{prefix}.{count_string}.{suffix}", type)
            result.add_field(field)
            result.set_value(field, value)

        count: int = 0
        width: int = len(str(len(self)))
        for obj in self:
            count += 1
            count_string: str = str(count + offset).rjust(width, "0")

            # Metadata
            for key, value in obj.metadata.items():
                # Get the datatype of the meta-data
                type: DataType = DataType.UNKNOWN
                if isinstance(value, Real):
                    type = DataType.NUMERIC
                elif isinstance(value, bool):
                    type = DataType.BOOLEAN
                elif isinstance(value, str):
                    type = DataType.STRING

                add_value(count_string, key, type, value)

            # Index
            if update_index:
                add_value(count_string, constants.KEY_INDEX, DataType.STRING, count_string)

            add_value(count_string, constants.KEY_X, DataType.NUMERIC, obj.x)  # X
            add_value(count_string, constants.KEY_Y, DataType.NUMERIC, obj.y)  # Y
            add_value(count_string, constants.KEY_WIDTH, DataType.NUMERIC, obj.width)  # Width
            add_value(count_string, constants.KEY_HEIGHT, DataType.NUMERIC, obj.height)  # Height
            add_value(count_string, constants.KEY_LOCATION, DataType.STRING, obj.get_location())  # Location

            # Polygon
            if obj.has_polygon():
                add_value(count_string, constants.KEY_POLY_X, DataType.STRING, ",".join(map(str, obj.get_polygon_x())))
                add_value(count_string, constants.KEY_POLY_Y, DataType.STRING, ",".join(map(str, obj.get_polygon_y())))

        # Count
        field: Field = Field(f"{prefix}.{constants.KEY_COUNT}", DataType.NUMERIC)
        result.add_field(field)
        result.set_value(field, len(self))

        return result

    @classmethod
    def from_report(cls, report: Report, prefix: str = "Object.") -> "LocatedObjects":
        """
        Retrieves all object from the report.

        :param report:  The report to process.
        :param prefix:  The prefix to look for.
        :return:        The objects found.
        """
        # Make sure the prefix ends in a dot
        if not prefix.endswith("."):
            prefix = prefix + "."

        result: LocatedObjects = LocatedObjects()
        fields: List[AbstractField] = report.get_fields()

        # Group fields
        groups: Dict[str, List[AbstractField]] = {}
        for field in fields:
            if field.name.startswith(prefix):
                current: str = field.name[:field.name.rindex(".")]
                if current not in groups:
                    groups[current] = []
                groups[current].append(field)

        # Process grouped fields
        for group, group_fields in groups.items():
            # Meta-data
            meta = {}
            if len(group) <= len(prefix):
                continue
            meta[constants.KEY_INDEX] = group[len(prefix):]
            for field in group_fields:
                fname = field.name[len(group) + 1:]
                if fname in [constants.KEY_X, constants.KEY_Y, constants.KEY_WIDTH, constants.KEY_HEIGHT]:
                    continue
                meta[field.name[field.name.rindex(".") + 1:]] = report.get_value(field)

            try:
                if (report.has_value(f"{group}.{constants.KEY_X}") and
                        report.has_value(f"{group}.{constants.KEY_Y}") and
                        report.has_value(f"{group}.{constants.KEY_WIDTH}") and
                        report.has_value(f"{group}.{constants.KEY_HEIGHT}")):
                    x: int = round(report.get_real_value(f"{group}.{constants.KEY_X}"))
                    y: int = round(report.get_real_value(f"{group}.{constants.KEY_Y}"))
                    width: int = round(report.get_real_value(f"{group}.{constants.KEY_WIDTH}"))
                    height: int = round(report.get_real_value(f"{group}.{constants.KEY_HEIGHT}"))
                    obj: LocatedObject = LocatedObject(x, y, width, height, **meta)
                    result.append(obj)

                    # Polygon
                    if (report.has_value(f"{group}.{constants.KEY_POLY_X}") and
                            report.has_value(f"{group}.{constants.KEY_POLY_Y}")):
                        obj.metadata[constants.KEY_POLY_X] = report.get_string_value(f"{group}.{constants.KEY_POLY_X}")
                        obj.metadata[constants.KEY_POLY_Y] = report.get_string_value(f"{group}.{constants.KEY_POLY_Y}")
            except Exception:
                # Ignored
                pass

        return result

    # PyCharm doesn't seem to be able to work out the typing for this
    def __iter__(self) -> Iterator[LocatedObject]:
        return super().__iter__()



"""
Types to do with code representations.
"""
from typing import Dict, Tuple, Callable, Any

# The types of a code-representation
ImportDict = Dict[str, str]
CodeRepresentation = Tuple[ImportDict, str]

# The type of the code_repr function
CodeReprFunction = Callable[[Any], CodeRepresentation]



from typing import Dict, Generic, TypeVar, IO

from .._Serialiser import Serialiser
from ._IntSerialiser import IntSerialiser

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class DictSerialiser(
    Serialiser[Dict[KeyType, ValueType]],
    Generic[KeyType, ValueType]
):
    def __init__(
            self,
            key_serialiser: Serialiser[KeyType],
            value_serialiser: Serialiser[ValueType],
            length_serialiser: Serialiser[int] = IntSerialiser(signed=False),
    ):
        super().__init__()

        self._key_serialiser: Serialiser[KeyType] = key_serialiser
        self._value_serialiser: Serialiser[ValueType] = value_serialiser
        self._length_serialiser: Serialiser[int] = length_serialiser

    def _check(
            self,
            obj: Dict[KeyType, ValueType]
    ):
        # Must be a dictionary
        if not isinstance(obj, dict):
            raise ValueError(f"{DictSerialiser.__class__.__name__} serialises dictionaries")

    def _serialise(
            self,
            obj: Dict[KeyType, ValueType],
            stream: IO[bytes]
    ):
        # Serialise the size of the dictionary
        self._length_serialiser.serialise(
            len(obj),
            stream
        )

        # Serialise each key/value pair in turn
        for key, value in obj.items():
            self._key_serialiser.serialise(key, stream)
            self._value_serialiser.serialise(value, stream)

    def _deserialise(
            self,
            stream: IO[bytes]
    ) -> Dict[KeyType, ValueType]:
        # Create a dict to accumulate the key/value pairs as they are deserialised
        result = {}

        # Deserialise the keys and values
        for _ in range(self._length_serialiser.deserialise(stream)):
            key = self._key_serialiser.deserialise(stream)
            value = self._value_serialiser.deserialise(stream)
            result[key] = value

        return result



"""
Package for working with sets, subsets and selection.
"""
from ._number_of_permutations import number_of_permutations
from ._number_of_subsets import number_of_subsets
from ._subset_number_to_subset import subset_number_to_subset
from ._subset_to_subset_number import subset_to_subset_number



import logging


class LoggingMixin:
    """
    Mixin class for adding logging to objects.
    """
    @property
    def logger(self) -> logging.Logger:
        cls = type(self)
        return logging.getLogger(cls.__module__ + "." + cls.__name__)



import os


def get_config_dir() -> str:
    """
    Gets the directory in which to store app configuration. Based on
    code from https://stackoverflow.com/a/3250952.

    On Windows, %APPDATA%\\
    On Linux, $XDG_CONFIG_HOME/

    If the relevant environment variable for the platform is not set,
    defaults to ~/.config/

    :return:    The directory string.
    """
    # Get the system app configuration standard location
    if 'APPDATA' in os.environ:
        return os.environ['APPDATA']
    elif 'XDG_CONFIG_HOME' in os.environ:
        return os.environ['XDG_CONFIG_HOME']
    else:
        return os.path.join(os.environ['HOME'], '.config')



"""
Module for typing-related types, methods, etc.
"""
from typing import TypeVar, Callable

# Standard unbounded type variable
T = TypeVar("T")

# Type for predicate functions
Predicate = Callable[[T], bool]



"""
Module containing tools for creating a standard root logger, for
when you don't want to think about how to add logging to your application.
"""
import logging
import sys
from typing import Callable

# The names for the default handlers
DEBUG_HANDLER_NAME = "debug_handler"
INFO_HANDLER_NAME = "info_handler"
WARNING_HANDLER_NAME = "warning_handler"
ERROR_HANDLER_NAME = "error_handler"


def exact_level_filter(level: int) -> Callable[[logging.LogRecord], bool]:
    """
    Creates a filter function which only allows records with a
    specified level though.

    :param level:   The level to filter for.
    :return:        The filter function.
    """
    def filter_function(record: logging.LogRecord) -> bool:
        return record.levelno == level

    return filter_function


def create_standard_handler(std_out: bool,
                            name: str,
                            level: int,
                            format_string: str,
                            exact_level: bool = True) -> logging.StreamHandler:
    """
    Creates a standard handler for printing log messages to
    either std-out or std-err.

    :param std_out:         Whether std-out should be used for printing. If False,
                            std-err is used instead.
    :param name:            The handler's name.
    :param level:           The level of message the handler should handle.
    :param format_string:   The format string for the log messages, in '{' style.
    :param exact_level:     Whether the handler should only handle records of an
                            exact level.
    :return:                The handler.
    """
    # Configure the handler
    handler = logging.StreamHandler(sys.stdout if std_out else sys.stderr)
    handler.name = name
    handler.setLevel(level)
    if exact_level:
        handler.addFilter(exact_level_filter(level))

    # Add the message format
    formatter = logging.Formatter(format_string, style="{")
    formatter.default_msec_format = "%s.%03d"
    handler.setFormatter(formatter)

    return handler


def create_standard_application_root_logger() -> logging.Logger:
    """
    Sets up a standard root logger and returns it. If the root
    logger has already been configured, issues a warning but leaves
    the logger unchanged.

    :return:        The root logger.
    """
    # Get the root logger
    logger = logging.getLogger()

    # If the logger already has been configured, warn and abort
    if len(logger.handlers) > 0:
        logger.warning("Attempted to reconfigure root logger")
        return logger

    # Create a debug handler to print debug info to std-out
    debug_handler = create_standard_handler(
        True,
        DEBUG_HANDLER_NAME,
        logging.DEBUG,
        "{levelname:8} {created:023} - {name} - {lineno} - {funcName} - {pathname}\n"
        "{message}"
    )

    # Create an info handler to print info messages to std-out
    info_handler = create_standard_handler(
        True,
        INFO_HANDLER_NAME,
        logging.INFO,
        "{levelname:8} {asctime:23} - {name} - {message}"
    )

    # Create a warning handler to print warning messages to std-err
    warning_handler = create_standard_handler(
        False,
        WARNING_HANDLER_NAME,
        logging.WARNING,
        "{levelname:8} {asctime:23} - {name} - {message}"
    )

    # Create a final handler to print all error/critical messages to std-err
    error_handler = create_standard_handler(
        False,
        ERROR_HANDLER_NAME,
        logging.ERROR,
        "{levelname:8} {asctime:23} - {name} - {message}",
        False
    )

    # Add the handlers to the logger
    logger.addHandler(debug_handler)
    logger.addHandler(info_handler)
    logger.addHandler(warning_handler)
    logger.addHandler(error_handler)

    # Set the main logger to warning level
    logger.setLevel(logging.WARNING)

    return logger



import javaproperties

from ._Report import Report
from . import constants


def save(report: Report, filename: str, encoding: str = "utf-8"):
    """
    Saves a report to disk.

    :param report:      The report to save.
    :param filename:    The filename to save the report under.
    :param encoding:    The encoding to use.
    """
    # Create a Java properties dictionary
    properties = {constants.PROPERTY_PARENTID: str(report.database_id)}

    # Add the fields and their types as properties
    for field in report:
        name = str(field)
        properties[name] = str(report.get_value(field))
        properties[name + constants.DATATYPE_SUFFIX] = field.datatype.to_display()

    # Write the properties to disk
    with open(filename, 'w', encoding=encoding) as file:
        print(javaproperties.to_comment(javaproperties.java_timestamp()), file=file)
        print(javaproperties.to_comment("Simple report format (= Java properties file format)"), file=file)
        javaproperties.dump(properties, file, timestamp=False, sort_keys=True)



import gzip
import os
from typing import Callable, TypeVar, Iterator

# Generic type parameter
T = TypeVar("T")


def get_open_func(filename):
    """
    Gets the open function to use to open the given filename.

    :param filename:    The file to open.
    :return:            The open function, and the read-mode flag to use.
    """

    # Select the open function based on the filename
    if filename.endswith('.gz'):
        return gzip.open, 'rt'
    else:
        return open, 'r'


def load_dir(load_function: Callable[[str], T],
             path: str,
             filename_predicate: Callable[[str], bool],
             recurse: bool = False) -> Iterator[T]:
    """
    Loads all files from the given path that match the filename predicate,
    using the given loading function.

    :param load_function:       The loading function to use to load the files.
    :param path:                The path to the directory to load from.
    :param filename_predicate:  A function which returns True if the file should be loaded,
                                and False if it should not be.
    :param recurse:           Whether sub-directories should be loaded as well.
    :return:                    An iterator over the loaded files.
    """
    # Make sure a valid path was provided
    if not os.path.isdir(path):
        raise ValueError("'" + path + "' is not a directory")

    # Loop through all files/directories in the given directory
    for name in os.listdir(path):
        # Get the full name
        full_name: str = os.path.join(path, name)

        # If we are recursing through directories, do that
        if os.path.isdir(full_name) and recurse:
            for loaded in load_dir(load_function, full_name, filename_predicate, recurse):
                yield loaded

        # Otherwise load the file if the name matches the predicate
        elif os.path.isfile(full_name) and filename_predicate(full_name):
            yield load_function(full_name)



from argparse import ArgumentParser

from ..meta import instanceoptionalmethod, non_default_kwargs
from .util import LoggingArgumentParser


class ArgumentParserConfigurer:
    """
    Interface for classes which can configure an argument parser.
    """
    @instanceoptionalmethod
    def configure_parser(self, parser: ArgumentParser):
        """
        Configures the parser.

        :param parser:  The parser to configure.
        """
        raise NotImplementedError(ArgumentParserConfigurer.configure_parser.__qualname__)

    @instanceoptionalmethod
    def get_configured_parser(self, *,
                              prog=...,
                              usage=...,
                              description=...,
                              epilog=...,
                              parents=...,
                              formatter_class=...,
                              prefix_chars=...,
                              fromfile_prefix_chars=...,
                              argument_default=...,
                              conflict_handler=...,
                              add_help=...,
                              allow_abbrev=...,
                              logging: bool = True) -> ArgumentParser:
        """
        Creates a configured parser.

        TODO: Parameter typing and descriptions
        TODO: Remove unsupported parameters e.g. prefix_chars
        TODO: Caching?

        :return:    The configured parser.
        """
        # Get the parser class based on the logging flag
        parser_class = LoggingArgumentParser if logging else ArgumentParser

        # Create a parser
        parser = parser_class(**non_default_kwargs(ArgumentParserConfigurer.get_configured_parser, locals()))

        # Configure it
        self.configure_parser(parser)

        return parser



from typing import Union, Tuple, Pattern, Iterable, List

import re

from ..sequence import normalise_index, flatten

SELECTION_TYPE = Union[int,  # Selects a column by index
                       Tuple[int, int],  # Selects a half-open range of columns by index
                       str,  # Selects columns by name using regex matching
                       Pattern,  # Selects columns by name using regex matching
                       Iterable['SELECTION_TYPE']]  # Any iterable of any of the above


class NamedColumnSelection:
    """
    Class representing a selection of columns from a table-like structure
    with named columns.
    """
    def __init__(self, selection: SELECTION_TYPE):
        self.selection: SELECTION_TYPE = selection

    def apply(self, column_names: List[str]) -> List[int]:
        return self.select(self.selection, column_names)

    @staticmethod
    def select(selection: SELECTION_TYPE, column_names: List[str]) -> List[int]:
        """
        Selects the indices of a set of columns.

        :param selection:       The selection criteria.
        :param column_names:    The names of the columns.
        :return:                The list of column indices selected.
        """
        # Calculate the number of columns
        num_columns: int = len(column_names)

        # Process integer selections by normalisation
        if isinstance(selection, int):
            # Return the normalised selection as a list of itself
            return [normalise_index(selection, num_columns)]

        # Process tuple selections as a half-open range of indices
        elif isinstance(selection, tuple):
            # Get the normalised start and end indices
            start = normalise_index(selection[0], num_columns)
            end = normalise_index(selection[1], num_columns)

            # If end is before start, return the indices in reverse order
            step = -1 if end < start else 1

            # Return the half-open range between start and stop, as a list
            return [i for i in range(start, end, step)]

        # Process strings and patterns by regex matching column names
        elif isinstance(selection, str) or isinstance(selection, Pattern):
            # Return any columns that match by name
            return [i
                    for i, column_name in enumerate(column_names)
                    if re.search(selection, column_name)]

        # Process lists by recurse-and-flatten
        elif isinstance(selection, list):
            sub_selections = [NamedColumnSelection.select(sub_selection, column_names)
                              for sub_selection in selection]

            return flatten(sub_selections)



from argparse import _AppendAction

# Capture the built-in type class as 'type' is an argument to __init__
_type = type


class ConcatAction(_AppendAction):
    """
    Similar to argparse's "append" action but concatenates
    the lists instead of nesting them. Also only returns the
    default value if no option is supplied, rather than
    concatenating into default a la "append".
    """
    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=None,
                 choices=None,
                 required=False,
                 help=None,
                 metavar=None):
        # Default default is an empty list
        if default is None:
            default = []

        # Can't have a default for a required option
        elif required:
            raise ValueError(f"Can't set a default for a required option")

        # Make sure default is a list
        if not _type(default) is list:
            raise TypeError(f"Default value for {ConcatAction.__qualname__} must be "
                            f"a list, but got a {type(default).__qualname__}")

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)
        items = [] if items is self.default else list(items)
        items += values if isinstance(values, list) else [values]
        setattr(namespace, self.dest, items)



from abc import abstractmethod


class SuffixOnlyField:
    """
    Interface for fields that only have a suffix.
    """
    @abstractmethod
    def is_compound(self) -> bool:
        pass

    @abstractmethod
    def get_suffix(self) -> str:
        pass



"""
Package for adding dynamic defaults to functions. Dynamic defaults
are default values for parameters which return a new object on every call,
rather than reusing the same object each call ("static" default).

Usage:
@with_dynamic_defaults
def my_func(a=dynamic_default(MyClass)):
    ...
"""
from ._dynamic_default import dynamic_default
from ._with_dynamic_defaults import with_dynamic_defaults



"""
Package for reading in .spec files.

Typical usage:
from wai.common.file import spec

for spectrum in spec.reader('filename.spec'):
    process_spectrum(spectrum)

OR

spectra = spec.read_all('filename.spec')
"""
from ._Field import Field, DATATYPE_BOOLEAN, DATATYPE_NUMERIC, DATATYPE_STRING, DATATYPE_UNKNOWN
from ._Report import Report, NO_ID
from ._SpecFileReader import read_all, reader, read_spectra, read_spectrum
from ._Spectrum import Spectrum
from ._SpectrumPoint import SpectrumPoint



import os
from multiprocessing.pool import Pool
from typing import IO, Callable, List, Any, Union

from .._FileReader import FileReader
from ._error import ARFFError
from ._ARFFFile import ARFFFile, DATA_TYPE, ROW_TYPE
from ._Attribute import Attribute
from . import constants
from ._load_helper import read_till_found, line_starts_with, is_comment_line, is_whitespace_only, remove_keyword, \
    consume, NAME_PATTERN, UnrecognisedContentError, remove_quotes, QUOTES

# The number of rows to process as a batch in parallel mode
PARALLEL_BATCH_SIZE = 64


class ARFFFileReader(FileReader[str, ARFFFile, "ARFFFileReader"]):
    """
    Reader for ARFF files.
    """
    def __init__(self, parallel: bool = True, use_csv: bool = True):
        super().__init__()

        self.parallel: bool = parallel
        self.use_csv: bool = use_csv

    def _load(self, file: IO[str]) -> ARFFFile:
        # Get the relation and attributes
        relation = get_relation_section(file)
        attributes = get_attribute_section(file)

        # Select the data-reading method (serial or parallel, csv or not)
        data_section_method: Callable[[IO[str], List[Attribute]], DATA_TYPE] = \
            get_data_section_parallel_csv if self.parallel and self.use_csv else \
            get_data_section_parallel if self.parallel else \
            get_data_section_csv if self.use_csv else \
            get_data_section

        # Get the data
        data = data_section_method(file, attributes)

        # Return the ARFF file
        return ARFFFile(relation, attributes, data)


def get_relation_section(file: IO[str]) -> str:
    """
    Gets the relation name from the ARFF file.

    :param file:    The file we are processing.
    :return:        The relation name.
    """
    # Get the line starting with the @relation keyword
    line = read_till_found(file, {constants.RELATION_SECTION_KEYWORD})

    # Parse the line
    relation = parse_relation_line(line)

    # Return the header
    return relation


def get_attribute_section(file: IO[str]) -> List[Attribute]:
    """
    Gets the list of attributes in the ARFF file.

    :param file:            The ARFF file being processed.
    :return:                The list of attributes in the file.
    """
    # Create the empty list of attributes
    attributes: List[Attribute] = []

    # Process attribute and stop on data keywords
    stopwords = {constants.ATTRIBUTE_SECTION_KEYWORD, constants.DATA_SECTION_KEYWORD}

    # Keep reading attributes until we find the @data keyword
    line = read_till_found(file, stopwords)
    while not line_starts_with(line, constants.DATA_SECTION_KEYWORD):
        # Parse the @attribute line
        attribute: Attribute = Attribute.from_string(line)

        # Add it to the list
        attributes.append(attribute)

        # Read the next attribute or data line
        line = read_till_found(file, stopwords)

    # Return the attributes
    return attributes


def get_data_section(file: IO[str], attributes: List[Attribute]) -> List[List[Any]]:
    """
    Gets the data from the data section of the ARFF file.

    :param file:        The ARFF file we are processing.
    :param attributes:  The attributes of the ARFF file.
    :return:            The list of data rows.
    """
    # Create the empty data list
    data = []

    # Keep reading lines until we run out of file
    line = None
    while line != '':
        # Read the next line
        line = file.readline()

        # If it was just whitespace/comments, skip it
        if is_whitespace_only(line) or is_comment_line(line):
            continue

        # Parse the data row
        row = parse_data_row_line(line, attributes)

        # Append the row to the data
        data.append(row)

    # Return the data
    return data


def get_data_section_parallel(file: IO[str], attributes: List[Attribute]) -> DATA_TYPE:
    """
    Gets the data from the data section of the ARFF file.

    :param file:        The ARFF file we are processing.
    :param attributes:  The attributes of the ARFF file.
    :return:            The list of data rows.
    """
    # Keep reading lines until we run out of file
    lines = [line for line in file if not (is_whitespace_only(line) or is_comment_line(line))]

    # Parse the lines in parallel
    with Pool(initializer=pool_initialiser_parallel, initargs=(attributes,)) as pool:
        return pool.map(parse_data_row_line_parallel, lines, PARALLEL_BATCH_SIZE)


def get_data_section_parallel_csv(file: IO[str], attributes: List[Attribute]) -> DATA_TYPE:
    """
    Gets the data from the data section of the ARFF file.
    Operates in a parallel manner and uses csv to split
    the lines of the file.

    :param file:        The ARFF file we are processing.
    :param attributes:  The attributes of the ARFF file.
    :return:            The list of data rows.
    """
    # Keep reading lines until we run out of file
    lines = [line for line in file if not (is_whitespace_only(line) or is_comment_line(line))]

    # Get the number of sub-processes we can employ
    pool_size: int = os.cpu_count()

    # Calculate the number of lines each sub-process should handle
    work_unit_size = len(lines) // pool_size

    # Divide the lines into the unit of work for each sub-process
    work_units: List[List[str]] = [lines[i * work_unit_size: (i+1) * work_unit_size] for i in range(pool_size - 1)]
    work_units.append(lines[(pool_size - 1) * work_unit_size:])

    # Process the lines in parallel
    import itertools
    with Pool(processes=pool_size, initializer=pool_initialiser_parallel, initargs=(attributes,)) as pool:
        return [row for row in itertools.chain(*pool.imap(parse_data_rows_csv_parallel, work_units))]


def get_data_section_csv(file: IO[str], attributes: List[Attribute]) -> DATA_TYPE:
    """
    Gets the data from the data section of the ARFF file.
    Uses csv to split the data lines.

    :param file:        The ARFF file we are processing.
    :param attributes:  The attributes of the ARFF file.
    :return:            The list of data rows.
    """
    # Keep reading lines until we run out of file
    lines = [line for line in file if not (is_whitespace_only(line) or is_comment_line(line))]

    # Parse the lines with csv and return the data
    return parse_data_rows_csv(lines, attributes)


def parse_data_rows_csv(lines: List[str], attributes: List[Attribute]) -> DATA_TYPE:
    """
    Parses a list of lines from the data section using csv.

    :param lines:           The lines to parse.
    :param attributes:      The attributes of the ARFF file.
    :return:                The parsed data.
    """
    # Create a CSV reader to split the lines
    import csv
    csv_reader = csv.reader(lines, doublequote=False, escapechar='\\', skipinitialspace=True, strict=True)

    # Calculate the number of elements that should be in each row
    num_attributes = len(attributes)

    # Initialise the output data to the correct size
    output = [[None] * num_attributes for _ in range(len(lines))]

    # Process each row
    for j, row in enumerate(csv_reader):
        # Make sure the line wasn't too long or too short
        if len(row) != num_attributes:
            raise DataSizeMismatchError(num_attributes, lines[j])

        # Convert each value
        for i in range(num_attributes):
            # Add the value to the row
            output[j][i] = attributes[i].parse_string(row[i])

    return output


def pool_initialiser_parallel(attributes: List[Attribute]):
    """
    Pool initialisation method for when using parallel data reading.

    :param attributes:  The attributes of the ARFF file.
    :return:            Nothing.
    """
    # Make the attributes global so parse_data_row_line_parallel can access them.
    global attributes_global
    attributes_global = attributes


def parse_data_row_line_parallel(line: str) -> ROW_TYPE:
    """
    Parallel implementation of parse_data_row_line.

    :param line:        The line of data to parse.
    :return:            A list of data values.
    """
    # Defers to normal implementation by accessing attributes from global
    return parse_data_row_line(line, attributes_global)


def parse_data_rows_csv_parallel(lines: List[str]) -> DATA_TYPE:
    """
    Parallel implementation of parse_data_rows_csv.

    :param lines:       The lines of data to parse.
    :return:            A list of data values.
    """
    # Defers to normal implementation by accessing attributes from global
    return parse_data_rows_csv(lines, attributes_global)


def parse_relation_line(line: str) -> str:
    """
    Gets the name of the relation from the line of the ARFF file
    beginning with the @relation keyword.

    :param line:    The line beginning with the @relation keyword.
    :return:        The name of the relation.
    """
    # Save the original line for error messages
    original_line = line

    # Remove the keyword and any following whitespace
    line = remove_keyword(line, constants.RELATION_SECTION_KEYWORD).lstrip()

    # Extract the relation name
    relation, line = consume(line, NAME_PATTERN)

    # Make sure we got a name back
    if relation is None:
        raise RelationNameNotFoundError(original_line)

    # Make sure the entire line has been consumed
    if not is_whitespace_only(line):
        raise UnrecognisedContentError(line, original_line)

    # Return the relation name
    return remove_quotes(relation)


def parse_data_row_line(line: str, attributes: List[Attribute]) -> ROW_TYPE:
    """
    Parses a line of data from the @data section of the ARFF file.

    :param line:            The line of data to parse.
    :param attributes:      The attributes of the ARFF file.
    :return:                A list of data values.
    """
    # Save the original line for error messages
    original_line = line

    # Get the number of expected values
    num_attributes = len(attributes)

    # Create the data row with all missing values
    row = [None] * num_attributes

    # Keep extracting values until we have enough
    i = 0
    j = 0
    line_length = len(line)
    while i < num_attributes:
        # Make sure there is string left to consume
        if j == line_length:
            break

        # Advance through any leading spaces
        while line[j] == ' ':
            j += 1

        # Advance through any quoted section
        if line[j] in QUOTES:
            # Advance till the next matching quote
            quote = line[j]
            j += 1
            start = j
            while line[j] != quote:
                j += 1
            end = j
            j += 2
        else:
            # Advance to the next delimiter
            start = j
            while line[j] not in {'\t', ',', '\n'}:
                j += 1
            end = j
            j += 1

        value = line[start:end]

        # Convert the value and add it to the output row
        row[i] = attributes[i].parse_string(value)

        # Increment the index
        i += 1

    # Make sure the line wasn't too long or too short
    if j != line_length or i != num_attributes:
        raise DataSizeMismatchError(num_attributes, original_line)

    return row


def parse_date_format(string: str) -> str:
    """
    Parses the date-format string for date-type attributes.
    TODO: Implement. csterling

    :param string:      The date-format string to parse.
    :return:            Currently, just the input.
    """
    return string


class RelationNameNotFoundError(ARFFError):
    """
    Exception for when the name of the relation cannot be found.
    """
    def __init__(self, line: str):
        super().__init__("Couldn't parse relation name from: " + line)


class DataSizeMismatchError(ARFFError):
    """
    Exception when there are too many or too few values in a given data row.
    """
    def __init__(self, num_attributes: int, line: str):
        super().__init__("Wrong number of values (require " + str(num_attributes) + ") in: " + line)



"""
Package for dealing with options in the context of a specific class/instance.
"""
from ._OptionClassContext import OptionClassContext
from ._OptionContext import OptionContext
from ._OptionInstanceContext import OptionInstanceContext



import io
from typing import Dict, Any, IO

import javaproperties

from .._functions import get_open_func
from ._Report import Report
from . import constants
from ._DataType import DataType
from ._Field import Field


def loadf(filename: str,
          encoding: str = 'utf-8') -> Report:
    """
    Reads a report file from the file given by name.

    :param filename:    The name of the file to read.
    :param encoding:    The encoding of the file.
    :return:            The loaded report file.
    """
    # Select the open method based on the filename
    open_func, mode = get_open_func(filename)

    # Parse the file
    with open_func(filename, mode, encoding=encoding) as file:
        return load(file)


def loads(string: str) -> Report:
    """
    Loads a report from a string.

    :param string:      Text in report file format.
    :return:            The report.
    """
    return load(io.StringIO(string))


def load(file: IO[str]) -> Report:
    """
    Loads a report file from an IO stream.

    :param file:    The stream to read the report from.
    :return:        The report.
    """
    # Load the report file as a Java properties file
    properties: Dict[str, Any] = javaproperties.load(file)

    # Parse the properties into a report
    return from_properties(properties)


def from_properties(properties: Dict[str, Any]) -> Report:
    """
    Parses the properties and generates a report object from it.

    :param properties:  The properties to generate the report from.
    :return:            The report.
    """
    # Create the result object
    report: Report = Report()

    # Parse the properties
    for property_name, property_value in properties.items():
        # Data-type properties are only used for the type of their corresponding property
        if property_name.endswith(constants.DATATYPE_SUFFIX):
            continue

        # Special handling for the parent ID field
        if property_name == constants.PROPERTY_PARENTID:
            report.database_id = int(property_value)

        # Get the datatype for this property
        datatype_property_name: str = property_name + constants.DATATYPE_SUFFIX
        datatype_property_value: str = DataType.UNKNOWN.value
        if datatype_property_name in properties:
            datatype_property_value = properties[datatype_property_name]
        datatype: DataType = DataType.parse(datatype_property_value)
        if datatype is None:
            datatype = DataType.UNKNOWN

        # Convert and set the value
        report.set_value(Field(property_name, datatype), datatype.convert(property_value))

    return report



class SpectrumPoint:
    """
    Class representing a single data point in a spectrum.
    """
    def __init__(self, wave_number: float = -1.0, amplitude: float = -1.0):
        self.wave_number: float = wave_number
        self.amplitude: float = amplitude



from .._OptionHandler import OptionHandler
from ..options import Option
from ._OptionContext import OptionContext


class OptionInstanceContext(OptionContext):
    """
    Class which handles accessing an option in the context of a
    specific option handler instance.
    """
    def __init__(self, option: Option, context: OptionHandler):
        # The option must be owned by the instance's type (or super-type)
        if not isinstance(context, option.owner):
            raise TypeError("Context instance does not have access to this option")

        super().__init__(option, context)



from ..options import Option


class OptionContext:
    """
    Base class for option contexts.
    """
    def __init__(self, option: Option, context):
        self._option: Option = option
        self._context = context



"""
Utility functions for working with paths and files.
"""
import os


def ensure_path(path: str):
    """
    Checks if the given path exists, and creates it if it does not.

    :param path:    The path.
    """
    if path != '' and not os.path.exists(path):
        os.mkdir(path)



"""
Module for constants related to abstract base-classes.
"""
# The name of the method attribute that specifies it is abstract
ABSTRACT_METHOD_ATTRIBUTE: str = "__isabstractmethod__"

# The name of the class attribute that defines the abstract methods
# for a class
ABSTRACT_CLASS_ATTRIBUTE: str = "__abstractmethods__"



"""
Package for reading ADAMS report files.
"""
from ._AbstractField import AbstractField
from ._DataType import DataType
from ._Field import Field
from ._load import load, loadf, loads, from_properties
from ._PrefixField import PrefixField
from ._PrefixOnlyField import PrefixOnlyField
from ._RegularField import RegularField
from ._Report import Report
from ._save import save
from ._SuffixField import SuffixField
from ._SuffixOnlyField import SuffixOnlyField
